from __future__ import annotations

import ipaddress
import json
import logging
import enum
import shutil
import time
import random
from IPython.core.display_functions import display
from datetime import datetime
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from fabric_cf.orchestrator.swagger_client import (
        Slice as OrchestratorSlice,
        Sliver as OrchestratorSliver,
    )
    from fabrictestbed_extensions.fablib.fablib import FablibManager

from concurrent import futures
from ipaddress import IPv6Address, ip_address
from fabrictestbed.slice_editor import Node as FimNode

from fabrictestbed.slice_editor import ExperimentTopology
from fabrictestbed_extensions.fablib.interface import Interface
from fabrictestbed_extensions.fablib.network_service import NetworkService
from fabrictestbed_extensions.fablib.node import Node
from fabrictestbed_extensions.fablib.slice import Slice
from fabrictestbed_extensions.fablib.site import Site, Host

CAPERSTART = "./caper/caper.byte -q -p -e"
PCAPDIR = "/home/ubuntu/pcaps/"
LOCALP4DIR = "."
REMOTEWORKDIR = ".crease"
CREASEDIR = "fabrictestbed-extensions/fabrictestbed_extensions/fablib/crease"
MONITORURL = "https://transparnet.cs.iit.edu/~awolosewicz/dpdk-crease_monitor"
DPDKNAME = "dpdk-crease_monitor"
MONPROT = 0x6587

UBUNTU_IMAGES = ["default_ubuntu_20",
                 "default_ubuntu_22",
                 "default_ubuntu_24",
                 "crease_ubuntu_22",
                 "docker_ubuntu_20",
                 "docker_ubuntu_22"]

class MonNetData(enum.IntEnum):
    NODENAME = 0
    IFACENAME = 1
    MIFACENAME = 2
    ISSINK = 3
    PORTNUM = 4

class CrinkleAnalyzer(Node):

    default_image = "crease_ubuntu_22"
    default_cores = 4
    default_ram = 16
    default_disk = 500

    def __init__(
        self,
        slice: Slice,
        node: FimNode,
        validate: bool = False,
        raise_exception: bool = False,
    ):
        super().__init__(slice=slice, node=node, validate=validate, raise_exception=raise_exception)

    @staticmethod
    def new_node(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        avoid: list[str] = [],
        validate: bool = False,
        raise_exception: bool = False,
    ):
        """
        Not intended for API call.  See: CrinkleSlice.add_monitor()

        Creates a new Crinkle analyzer FABRIC node and returns a fablib node with the
        new node.

        :param slice: the fablib slice to build the new node on
        :type slice: Slice

        :param name: the name of the new node
        :type name: str

        :param site: the name of the site to build the node on
        :type site: str

        :param avoid: a list of node names to avoid
        :type avoid: List[str]

        :param validate: Validate node can be allocated w.r.t available resources
        :type validate: bool

        :param raise_exception: Raise exception in case of failure
        :type raise_exception: bool

        :return: a new fablib node
        :rtype: CrinkleAnalyzer
        """
        if site is None:
            [site] = slice.get_fablib_manager().get_random_sites(avoid=avoid)
        
        logging.info(f"Adding Crinkle Analyzer {name}, slice: {slice.get_name()}, site: {site}")

        analyzer = CrinkleAnalyzer(
            slice,
            slice.topology.add_node(name=name, site=site),
            validate=validate,
            raise_exception=raise_exception
        )

        analyzer.set_capacities(
            cores=CrinkleAnalyzer.default_cores,
            ram=CrinkleAnalyzer.default_ram,
            disk=CrinkleAnalyzer.default_disk
        )

        analyzer.set_image(CrinkleAnalyzer.default_image)

        analyzer.init_fablib_data()

        return analyzer
    
    @staticmethod
    def get_node(slice: Slice = None, node=None):
        """
        Returns a new Crinkle monitor node using existing FABRIC resources.

        :note: Not intended for API call.

        :param slice: the fablib slice storing the existing node
        :type slice: Slice

        :param node: the FIM node stored in this fablib node
        :type node: Node

        :return: a new fablib node storing resources
        :rtype: CrinkleMonitor
        """
        return CrinkleAnalyzer(slice, node)
    
class CrinkleMonitor(Node):

    default_image = "crease_ubuntu_22"
    default_cores = 2
    default_ram = 4
    default_disk = 10

    class MonitorData():
        def __init__(
            self,
            port_nums: int = 0,
            cmd_args: str = "",
            net_name: str = None,
            net_type: str = None,
            cnet_iface: Interface = None,
            iface_mappings: dict[str, tuple[str, Interface, bool, int]] = {},
            monitor_id: int = None
        ):
            self.port_nums = port_nums
            self.cmd_args = cmd_args
            self.net_name = net_name
            self.net_type = net_type
            self.cnet_iface = cnet_iface
            self.iface_mappings = iface_mappings
            self.monitor_id = monitor_id

    def __init__(
        self,
        slice: Slice,
        node: FimNode,
        validate: bool = False,
        raise_exception: bool = False,
    ):
        super().__init__(slice=slice, node=node, validate=validate, raise_exception=raise_exception)
        self.get_monitor_data()
        self.creation_data: list[tuple[str, str, str, bool, int]] = [] # see MonNetData

    @staticmethod
    def new_node(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        avoid: list[str] = [],
        validate: bool = False,
        raise_exception: bool = False,
    ):
        """
        Not intended for API call.  See: CrinkleSlice.add_monitor()

        Creates a new Crinkle monitor FABRIC node and returns a fablib node with the
        new node.

        :param slice: the fablib slice to build the new node on
        :type slice: Slice

        :param name: the name of the new node
        :type name: str

        :param site: the name of the site to build the node on
        :type site: str

        :return: a new fablib node
        :rtype: CrinkleMonitor
        """
        logging.info(f"Adding Crinkle Monitor {name}, slice: {slice.get_name()}, site: {site}")
        
        monitor = CrinkleMonitor(
            slice,
            slice.topology.add_node(name=name, site=site)
        )

        monitor.set_capacities(
            cores=CrinkleMonitor.default_cores,
            ram=CrinkleMonitor.default_ram,
            disk=CrinkleMonitor.default_disk
        )

        monitor.set_image(CrinkleMonitor.default_image)

        monitor.init_fablib_data()

        return monitor
    
    @staticmethod
    def get_node(slice: Slice = None, node=None):
        """
        Returns a new Crinkle monitor node using existing FABRIC resources.

        :note: Not intended for API call.

        :param slice: the fablib slice storing the existing node
        :type slice: Slice

        :param node: the FIM node stored in this fablib node
        :type node: Node

        :return: a new fablib node storing resources
        :rtype: CrinkleMonitor
        """
        return CrinkleMonitor(slice, node)
    
    def get_monitor_data(self):
        """
        Get monitor-specific data.
        """
        logging.info(f"{self.get_name()} get_monitor_data()")
        if "monitor_config" in self.get_user_data():
            data = self.get_user_data()["monitor_config"]
            #data_iface_mappings = self.get_user_data()["iface_mappings"]
            #logging.info(f"Retrieved monitor iface mappings as: {data_iface_mappings}")
            iface_mappings = {}
            for node_iface, (node_name, monitor_iface, is_sink, port_num) in data["iface_mappings"].items():
                iface_mappings[node_iface] = (
                    node_name, self.slice.get_interface(name=monitor_iface), is_sink, port_num
                )
            self.data = self.MonitorData(
                port_nums=data["port_nums"],
                cmd_args=data["cmd_args"],
                net_name=data["net_name"],
                net_type=data["net_type"],
                cnet_iface=self.slice.get_interface(data["cnet_iface"]),
                monitor_id=data["monitor_id"],
                iface_mappings=iface_mappings
            )
            logging.info(f"Retrieved monitor config as: {self.data.__dict__}")
        else:
            self.data = self.MonitorData()
            logging.info(f"Did not retrieve stored monitor data, initializing")
        
    def set_monitor_data(self):
        """
        Set monitor-specific data.
        """
        logging.info(f"{self.get_name()} set_monitor_data()")
        user_data = self.get_user_data()
        iface_mappings = {}
        for node_iface, (node_name, monitor_iface, is_sink, port_num) in self.data.iface_mappings.items():
            iface_mappings[node_iface] = (
                node_name, monitor_iface.get_name(), is_sink, port_num
            )
        data_dict = {
            "port_nums": self.data.port_nums,
            "cmd_args": self.data.cmd_args,
            "net_name": self.data.net_name,
            "net_type": self.data.net_type,
            "cnet_iface": self.data.cnet_iface.get_name(),
            "monitor_id": self.data.monitor_id,
            "iface_mappings": iface_mappings
        }
        logging.info(f"Writing monitor config to user data: {data_dict}")
        user_data["monitor_config"] = data_dict
        #logging.info(f"Writing monitor iface mappings to user data: {iface_mappings}")
        #user_data["iface_mappings"] = iface_mappings
        self.set_user_data(user_data=user_data)
    
class CrinkleSlice(Slice):
    def __init__(
            self,
            fablib_manager: FablibManager,
            name: str = None,
            user_only: bool = True,
            pcaps_dir: str = None,
            name_prefix: str = None,
            analyzer_name: str = None
    ):
        super().__init__(fablib_manager=fablib_manager, name=name, user_only=user_only)
        self.monitors: dict[str, CrinkleMonitor] = {}
        self.analyzer: CrinkleAnalyzer = None
        self.analyzer_name: str = analyzer_name
        self.cnets: dict[str, NetworkService] = {}
        self.analyzer_cnet: NetworkService = None
        self.analyzer_iface: Interface = None
        self.pcaps_dir = pcaps_dir
        self.prefix = name_prefix
        self.monitor_count = 0
        self.monitor_string = None
        self.all_interfaces = {}
        self.probe_id = 0
        self.do_allocate_hosts = True
        self.do_post_boot = True
        self.do_ptp_setup = True

    @staticmethod
    def new_slice(
            fablib_manager: FablibManager, 
            name: str = None,
            pcaps_dir: str = ".query_analysis_pcaps",
            name_prefix: str = "C"
        ):
        """
        Create a new crinkle slice
        :param fablib_manager:
        :param name:
        :param analyzer_name:
        :param cores:
        :param ram:
        :param disk:
        :param site:
        :param image:
        :param pcaps_dir:
        :return: CrinkleSlice
        """
        slice = CrinkleSlice(fablib_manager=fablib_manager, name=name, pcaps_dir=pcaps_dir,
                             name_prefix=name_prefix)
        slice.topology = ExperimentTopology()
        if fablib_manager:
            fablib_manager.cache_slice(slice_object=slice)
        return slice
    
    def get_crinkle_data(self, analyzer: CrinkleAnalyzer):
        """
        Get slice-wide crinkle data.
        """
        logging.info(f"get_crinkle_data()")
        if "crinkle_slice_config" in analyzer.get_user_data():
            data = self.analyzer.get_user_data()["crinkle_slice_config"]
            #data_iface_mappings = self.get_user_data()["iface_mappings"]
            #logging.info(f"Retrieved monitor iface mappings as: {data_iface_mappings}")
            self.analyzer_name = data["analyzer_name"]
            self.prefix = data["prefix"]
            self.monitor_string = data["monitor_string"]
            self.do_allocate_hosts = data["do_allocate_hosts"]
            self.do_post_boot = data["do_post_boot"]
            self.do_ptp_setup = data["do_ptp_setup"]
            logging.info(f"Retrieved crinkle slice config as:\n{self.analyzer_name}\n{self.prefix}\n{self.monitor_string}")
        else:
            logging.info(f"Did not retrieve stored crinkle slice config")
        
    def set_crinkle_data(self):
        """
        Set slice-wide crinkle data.
        """
        logging.info(f"set_crinkle_data()")
        user_data = self.analyzer.get_user_data()
        data_dict = {
            "analyzer_name": self.analyzer_name,
            "prefix": self.prefix,
            "monitor_string": self.monitor_string,
            "do_allocate_hosts": self.do_allocate_hosts,
            "do_post_boot": self.do_post_boot,
            "do_ptp_setup": self.do_ptp_setup
        }
        logging.info(f"Writing crinkle slice config to user data: {data_dict}")
        user_data["crinkle_slice_config"] = data_dict
        #logging.info(f"Writing monitor iface mappings to user data: {iface_mappings}")
        #user_data["iface_mappings"] = iface_mappings
        self.analyzer.set_user_data(user_data=user_data)
    
    @staticmethod
    def get_slice(
        fablib_manager: FablibManager,
        sm_slice: OrchestratorSlice = None,
        user_only: bool = True,
        pcaps_dir: str = ".query_analysis_pcaps",
        name_prefix: str = "C"
    ):
        """
        Not intended for API use. See FablibManager.get_crinkle_slice().

        Gets an existing crinkle fablib slice using a slice manager slice
        :param fablib_manager:
        :param sm_slice:
        :param user_only: True indicates return own slices; False indicates return project slices
        :type user_only: bool
        :param name_prefix: The prefix Crinkle will append its resources with, which should not prefix any other resources
        :type name_prefix: String
        :return: CrinkleSlice
        """
        logging.info("crinkleslice.get_slice()")
        slice = CrinkleSlice(fablib_manager=fablib_manager, name=sm_slice.name,
                             pcaps_dir=pcaps_dir, name_prefix=name_prefix)
        slice.sm_slice = sm_slice
        slice.slice_id = sm_slice.slice_id
        slice.slice_name = sm_slice.name
        slice.user_only = user_only
        if fablib_manager:
            fablib_manager.cache_slice(slice_object=slice)

        try:
            slice.update_topology()
        except Exception as e:
            logging.error(
                f"Slice {slice.slice_name} could not update topology: slice.get_slice"
            )
            logging.error(e, exc_info=True)

        try:
            slice.update_slivers()
        except Exception as e:
            logging.error(
                f"Slice {slice.slice_name} could not update slivers: slice.get_slice"
            )
            logging.error(e, exc_info=True)

        slice.analyzer = slice.get_analyzer(name=f"{name_prefix}_analyzer")
        slice.get_crinkle_data(slice.analyzer)
        analyzer_site = slice.analyzer.get_site()
        for net in slice.get_networks():
            if net.get_name().startswith(f"{name_prefix}_ananet_"):
                slice.cnets[net.get_site()] = net
        slice.analyzer_cnet = slice.cnets[analyzer_site]
        slice.analyzer_name = slice.analyzer.get_name()
        slice.analyzer_iface = slice.analyzer.get_interface(network_name=slice.analyzer_cnet.get_name())

        for node in slice.get_all_nodes():
            node_name: str = node.get_name()
            if node_name.startswith(f"{name_prefix}_monitor_"):
                monitor = slice.get_monitor(name=node_name)
                monitor.get_monitor_data()
                slice.monitors[monitor.data.net_name] = monitor
                slice.monitor_count += 1
        
        if slice.monitor_string == '':
            slice.reset_monitor_string()
        slice.set_crinkle_data()

        return slice
    
    def add_analyzer(
            self,
            site: str = None,
            cores: int = CrinkleAnalyzer.default_cores,
            ram: int = CrinkleAnalyzer.default_ram,
            disk: int = CrinkleAnalyzer.default_disk,
            instance_type: str = None,
            host: str = None,
            user_data: dict = {},
            avoid: list[str] = [],
            validate: bool = False,
            raise_exception: bool = False,
        ) -> CrinkleAnalyzer:
        """
        Creates a new Crinkle Analyzer node on this fablib slice.

        :param site: (Optional) Name of the site to deploy the node
            on.  Default to a random site.
        :type site: String

        :param cores: (Optional) Number of cores in the node.
            Default: 2 cores
        :type cores: int

        :param ram: (Optional) Amount of ram in the node.  Default: 8
            GB
        :type ram: int

        :param disk: (Optional) Amount of disk space n the node.
            Default: 10 GB
        :type disk: int

        :param instance_type:
        :type instance_type: String

        :param host: (Optional) The physical host to deploy the node.
            Each site has worker nodes numbered 1, 2, 3, etc.  Host
            names follow the pattern in this example of STAR worker
            number 1: "star-w1.fabric-testbed.net".  Default: unset
        :type host: String

        :param user_data
        :type user_data: dict

        :param avoid: (Optional) A list of sites to avoid is allowing
            random site.
        :type avoid: List[String]

        :param validate: Validate node can be allocated w.r.t available resources
        :type validate: bool

        :param raise_exception: Raise exception in case of Failure
        :type raise_exception: bool

        :return: a new Crinkle Analyzer node
        :rtype: CrinkleAnalyzer
        """

        analyzer = CrinkleAnalyzer.new_node(
            slice=self,
            name=f"{self.prefix}_analyzer",
            site=site,
            avoid=avoid,
            validate=validate,
            raise_exception=raise_exception
        )

        analyzer.init_fablib_data()

        user_data_working = analyzer.get_user_data()
        for k, v in user_data.items():
            user_data_working[k] = v
        analyzer.set_user_data(user_data_working)

        if instance_type:
            analyzer.set_instance_type(instance_type)
        else:
            analyzer.set_capacities(cores=cores, ram=ram, disk=disk)

        analyzer.set_image(CrinkleAnalyzer.default_image)

        if host:
            analyzer.set_host(host)

        self.nodes = None
        self.interfaces = {}

        if validate:
            status, error = self.get_fablib_manager().validate_node(node=analyzer)
            if not status:
                analyzer.delete()
                analyzer = None
                logging.warning(error)
                if raise_exception:
                    raise ValueError(error)

        self.analyzer = analyzer
        self.analyzer_name = analyzer.get_name()
        analyzer_site = self.analyzer.get_site()
        if analyzer_site not in self.cnets or self.cnets[analyzer_site] is None:
            self.cnets[analyzer_site] = self.add_l3network(name=f"{self.prefix}_ananet_{site}", type="IPv6")
        self.analyzer_cnet = self.cnets[analyzer_site]
        analyzer_iface = self.analyzer.add_component(model="NIC_Basic",
                                                     name=f"{self.prefix}_nic_cnet").get_interfaces()[0]
        analyzer_iface.set_mode("auto")
        self.analyzer_cnet.add_interface(analyzer_iface)
        # Import here due to circular import issues
        from fabrictestbed_extensions.fablib.fablib import FablibManager
        self.analyzer.add_route(subnet=FablibManager.FABNETV6_SUBNET, next_hop=self.analyzer_cnet.get_gateway())

        return analyzer
    
    def add_monitor(
            self,
            name: str,
            site: str = None,
            cores: int = CrinkleMonitor.default_cores,
            ram: int = CrinkleMonitor.default_ram,
            disk: int = CrinkleMonitor.default_disk,
            image: str = CrinkleMonitor.default_image,
            instance_type: str = None,
            host: str = None,
            user_data: dict = {},
            avoid: list[str] = [],
            validate: bool = False,
            raise_exception: bool = False,
            net_name: str = None
    ) -> CrinkleMonitor:
        """
        Not intended for API call.
        See: CrinkleSlice.add_monitored_l2network() or CrinkleSlice.add_monitored_l3network()
        Creates a new Crinkle monitor node.

        :param name: The name for the monitor node
        :type name: String
        :param site:
        :type site: String
        :param user_data
        :type user_data: dict
        :param net_name: The name of the network this will monitor
        :type net_name: String
        :return: a new monitor node
        :rtype: CrinkleMonitor
        """
        if self.analyzer is None:
            raise Exception(f"Analyzer must be created before adding monitors using add_analyzer()")
        
        monitor = CrinkleMonitor.new_node(
            slice=self,
            name=name,
            site=site,
        )

        monitor.init_fablib_data()

        user_data_working = monitor.get_user_data()
        for k, v in user_data.items():
            user_data_working[k] = v
        monitor.set_user_data(user_data_working)
        
        monitor.set_capacities(cores=cores, ram=ram, disk=disk)
        
        monitor.set_image(image)

        if host:
            monitor.set_host(host)

        self.nodes = None
        self.interfaces = {}
        
        if site not in self.cnets or self.cnets[site] is None:
            self.cnets[site] = self.add_l3network(name=f"{self.prefix}_net_{site}", type="IPv6")
        cnet = self.cnets[site]
        monitor_cnet_iface = monitor.add_component(model="NIC_Basic",
                                                   name=f"{self.prefix}_nic_cnet").get_interfaces()[0]
        monitor_cnet_iface.set_mode("auto")
        cnet.add_interface(monitor_cnet_iface)
        monitor.add_route(subnet=self.analyzer_cnet.get_subnet(), next_hop=cnet.get_gateway())
        monitor.data.cnet_iface = monitor_cnet_iface
        monitor.data.net_name = net_name
        monitor.data.monitor_id = self.monitor_count
        self.monitor_count += 1
        monitor.set_monitor_data()
        
        return monitor
    
    def get_analyzer(self, name:str) -> CrinkleAnalyzer:
        """
        Gets an analyzer from the CrinkleSlice by name.

        :param name: Name of the analyzer
        :type name: String
        :return: the analyzer node
        :rtype: CrinkleAnalyzer
        """
        try:
            return CrinkleAnalyzer.get_node(self, self.get_fim_topology().nodes[name])
        except Exception as e:
            logging.info(e, exc_info=True)
            raise Exception(f"Node not found: {name}")
    
    def get_monitor(self, name: str) -> CrinkleMonitor:
        """
        Gets a monitor from the CrinkleSlice by name.

        :param name: Name of the monitor
        :type name: String
        :return: the monitor node
        :rtype: CrinkleMonitor
        """
        try:
            return CrinkleMonitor.get_node(self, self.get_fim_topology().nodes[name])
        except Exception as e:
            logging.info(e, exc_info=True)
            raise Exception(f"Node not found: {name}")

    def add_monitored_l2network(
        self,
        name: str = None,
        interfaces: list[Interface] = [],
        type: str = None,
        subnet: ipaddress = None,
        gateway: ipaddress = None,
        user_data: dict = {},
        sinks: list[Interface] = [],
        host: str = None,
        site: str = None,
        cores: int = CrinkleMonitor.default_cores,
        ram: int = CrinkleMonitor.default_ram,
        disk: int = CrinkleMonitor.default_disk
    ) -> CrinkleMonitor:
        """
        Adds an L2 network similarly to Slice.add_l2network, but additionally adds a
        CrinkleMonitor node to the network which all traffic routes through. The
        CrinkleMonitor nodes will collect traffic information and update the
        CrinkleAnalyzer node. Additionally, certain functions will use the CrinkleMonitor
        nodes to emit or modify packets.

        

        :param name: the name of the network service
        :type name: String

        :param interfaces: a list of interfaces to build the network
            with
        :type interfaces: List[Interface]

        :param type: optional L2 network type "L2Bridge", "L2STS", or
            "L2PTP"
        :type type: String

        :param subnet:
        :type subnet: ipaddress

        :param gateway:
        :type gateway: ipaddress

        :param user_data
        :type user_data: dict

        :param sinks: A set of interfaces which should have any Crinkle packet trailers
            stripped before entering.
        :type sinks: list[Interface]

        :return: a new CrinkleMonitor
        :rtype: CrinkleMonitor
        """
        # Directly from NetworkService.__calculate_l2_nstype
        from fabrictestbed_extensions.fablib.facility_port import FacilityPort

        # if there is a basic NIC, WAN must be STS
        basic_nic_count = 0

        sites = set([])
        includes_facility_port = False
        facility_port_interfaces = 0
        for interface in interfaces:
            sites.add(interface.get_site())
            if isinstance(interface.get_node(), FacilityPort):
                includes_facility_port = True
                facility_port_interfaces += 1
            if interface.get_model() == "NIC_Basic":
                basic_nic_count += 1

        rtn_nstype = None
        if 1 >= len(sites) >= 0:
            rtn_nstype = NetworkService.network_service_map["L2Bridge"]
        elif len(sites) == 2:
            # Use L2STS when connecting two facility ports instead of L2PTP
            # L2PTP limitation for Facility Ports:
            # basically the layer-2 point-to-point server template applied is not popping
            # vlan tags over the MPLS tunnel between two facility ports.
            if (
                includes_facility_port and facility_port_interfaces < 2
            ) and not basic_nic_count:
                # For now WAN FacilityPorts require L2PTP
                rtn_nstype = NetworkService.network_service_map["L2PTP"]
            elif len(interfaces) >= 2:
                rtn_nstype = NetworkService.network_service_map["L2STS"]
        else:
            raise Exception(
                f"Invalid Network Service: Networks are limited to 2 unique sites. Site requested: {sites}"
            )
        type = str(rtn_nstype)
        if not site:
            site = interfaces[0].get_site()
        monitor = self.add_monitor(name=f"{self.prefix}_monitor_{name}",
                                   site=site,
                                   net_name=name,
                                   host=host,
                                   cores=cores,
                                   ram=ram,
                                   disk=disk)

        if type == "L2Bridge":
            for iface in interfaces:
                iface_node_name = iface.get_node().get_name()
                monitor_iface = monitor.add_component("NIC_Basic", f"{self.prefix}_nic_{iface_node_name}").get_interfaces()[0]
                monitor_iface.set_mode("manual")
                new_net = self.add_l2network(f"{self.prefix}_net_{name}_{iface_node_name}", [iface, monitor_iface], "L2Bridge", subnet, gateway, user_data)
                self.set_orig_net_name(new_net, name)
                monitor.data.net_type = type
                monitor.creation_data.append((iface_node_name, iface.get_name(), f"{self.prefix}_nic_{iface_node_name}", (iface in sinks), 0))
            self.monitors[name] = monitor
        monitor.set_monitor_data()
        return monitor
    
    @staticmethod
    def get_orig_net_name(net: NetworkService) -> str:
        user_data = net.get_user_data()
        if "crinkle_net_name" in user_data:
            orig_name = user_data["crinkle_net_name"]
            logging.info(f"Retrieved original network name for {net.get_name()} as {orig_name}")
            return orig_name
        else:
            logging.info(f"Did not retrieve original network name for {net.get_name()}")
            return ""

    @staticmethod
    def set_orig_net_name(net: NetworkService, name: str):
        logging.info(f"Storing orig name for network {net.get_name()} as {name}")
        user_data = net.get_user_data()
        user_data["crinkle_net_name"] = name
        net.set_user_data(user_data=user_data)

    def allocate_hosts(self):
        allocated = {}
        logging.info("Allocating Monitors and their connected Nodes to different worker hosts")
        sitenames_to_sites: dict[str, Site] = {}
        sitenames_to_hosts: dict[str, dict[str, Host]] = {}
        fablib = self.get_fablib_manager()
        fabresources = fablib.get_resources()
        validated_nodes: dict[str, bool] = {}

        for node in self.get_all_nodes():
            host_name = node.get_host()
            if host_name is None:
                validated_nodes[node.get_name()] = False
                continue
            allocated_comps = allocated.setdefault(host_name, {})
            site_name = node.get_site()
            site = sitenames_to_sites.setdefault(site_name, fabresources.get_site(site_name))
            hosts = sitenames_to_hosts.setdefault(site_name, site.get_hosts())
            host = hosts[host_name]
            if fablib._FablibManager__can_allocate_node_in_host(
                host=host, node=node, allocated=allocated_comps, site=site)[0]:
                validated_nodes[node.get_name()] = True
            else:
                raise Exception(f"Host {host_name} does not have the free resources to reserve Node {node.get_name()}")

        for _, monitor in self.monitors.items():
            logging.info(f"Allocating Monitor for network {monitor.data.net_name}")
            if monitor.data.net_type == "L2Bridge":
                endpoint1 = self.get_node(name=monitor.creation_data[0][0])
                endpoint2 = self.get_node(name=monitor.creation_data[1][0])
                endpoints = [endpoint1, endpoint2]
                site_name = endpoint1.get_site()
                site = sitenames_to_sites.setdefault(site_name, fabresources.get_site(site_name))
                hosts = sitenames_to_hosts.setdefault(site_name, site.get_hosts())
                hostlist = list(hosts.items())
                hostlist = sorted(hostlist)
                endhosts = {}
                for endpoint in endpoints:
                    endpoint_name = endpoint.get_name()
                    if validated_nodes[endpoint_name]:
                        logging.info(f"Node {endpoint_name} already allocated to {endpoint.get_host()}")
                        endhosts.setdefault(endpoint.get_host(), True)
                        continue
                    # TODO: instead, traverse hostlist in reverse
                    for host_name, host in hostlist[1:]:
                        allocated_comps = allocated.setdefault(host_name, {})
                        if fablib._FablibManager__can_allocate_node_in_host(
                            host=host, node=endpoint, allocated=allocated_comps, site=site)[0]:
                            endpoint.set_host(host_name=host_name)
                            endhosts.setdefault(host_name, True)
                            validated_nodes[endpoint_name] = True
                            logging.info(f'Node {endpoint_name} assigned to {host_name}')
                            break
                    if not validated_nodes[endpoint_name]:
                        raise Exception(f"Could not place node {endpoint_name} due to a lack of free workers. Please try another site.")
                monitor_name = monitor.get_name()
                if validated_nodes.get(monitor_name, False):
                    logging.info(f"Monitor {monitor_name} already allocated to {monitor.get_host()}")
                    continue
                for host_name, host in hostlist:
                    if host_name in endhosts:
                        continue
                    allocated_comps = allocated.setdefault(host_name, {})
                    if fablib._FablibManager__can_allocate_node_in_host(
                        host=host, node=monitor, allocated=allocated_comps, site=site)[0]:
                        monitor.set_host(host_name=host_name)
                        validated_nodes[monitor_name] = True
                        logging.info(f'Node {monitor_name} assigned to {host_name}')
                        break
                if not validated_nodes[monitor_name]:
                    raise Exception(f"Could not place monitor for network {monitor.data.net_name} due to a lack of free workers. Please try another site.")
        
        for node in self.get_all_nodes():
            node_name = node.get_name()
            if validated_nodes[node_name]:
                continue
            site_name = node.get_site()
            site = sitenames_to_sites.setdefault(site_name, fabresources.get_site(site_name))
            hosts = sitenames_to_hosts.setdefault(site_name, site.get_hosts())
            hostlist = list(hosts.items())
            hostlist = sorted(hostlist)
            for host_name, host in hostlist:
                allocated_comps = allocated.setdefault(host_name, {})
                if fablib._FablibManager__can_allocate_node_in_host(
                    host=host, node=node, allocated=allocated_comps, site=site)[0]:
                    node.set_host(host_name=host_name)
                    validated_nodes[node_name] = True
                    logging.info(f"Node {node_name} assigned to {host_name}")
                    break
            if not validated_nodes[node_name]:
                raise Exception(f"Could not place node {node_name} due to a lack of free workers. Please try another site.")

        self.do_allocate_hosts = False
        self.set_crinkle_data()
        logging.info(f"Hosts allocated")
        for host_name in allocated:
            site_name = host_name.split("-")[0].upper()
            site = sitenames_to_sites.setdefault(site_name, fabresources.get_site(site_name))
            hosts = sitenames_to_hosts.setdefault(site_name, site.get_hosts())
            host = hosts[host_name]
            allocated_comps = allocated[host_name]
            logging.info(f"{host_name}: CPU {allocated_comps['core']}/{host.get_core_available()} "
                            f"RAM {allocated_comps['ram']}/{host.get_ram_available()} "
                            f"DISK {allocated_comps['disk']}/{host.get_disk_available()}")
    
    def submit(
        self,
        wait: bool = True,
        wait_timeout: int = 1800,
        wait_interval: int = 20,
        progress: bool = True,
        wait_jupyter: str = "text",
        post_boot_config: bool = True,
        wait_ssh: bool = True,
        extra_ssh_keys: list[str] = None,
        lease_start_time: datetime = None,
        lease_end_time: datetime = None,
        lease_in_hours: int = None,
        validate: bool = False,
    ) -> str:
        """
        Similarly to Slice.submit(), submits a slice request to FABRIC.
        This version does extra work to support Crinkle resources, such as
        ensuring monitors are on different workers from the nodes they monitor.


        :param wait: indicator for whether to wait for the slice's resources to be active
        :type wait: bool

        :param wait_timeout: how many seconds to wait on the slice resources
        :type wait_timeout: int

        :param wait_interval: how often to check on the slice resources
        :type wait_interval: int

        :param progress: indicator for whether to show progress while waiting
        :type progress: bool

        :param wait_jupyter: Special wait for jupyter notebooks.
        :type wait_jupyter: str

        :param post_boot_config:
        :type post_boot_config: bool

        :param wait_ssh:
        :type wait_ssh: bool

        :param extra_ssh_keys: Optional list of additional SSH public keys to be installed in the slivers of this slice
        :type extra_ssh_keys: List[str]

        :param lease_start_time: Optional lease start time in UTC format: %Y-%m-%d %H:%M:%S %z.
                           Specifies the beginning of the time range to search for available resources valid for `lease_in_hours`.
        :type lease_start_time: datetime

        :param lease_end_time: Optional lease end time in UTC format: %Y-%m-%d %H:%M:%S %z.
                         Specifies the end of the time range to search for available resources valid for `lease_in_hours`.
        :type lease_end_time: datetime

        :param lease_in_hours: Optional lease duration in hours. By default, the slice remains active for 24 hours (1 day).
                               This parameter is only applicable during creation.
        :type lease_in_hours: int

        :param validate: Validate node can be allocated w.r.t available resources
        :type validate: bool

        :return: slice_id
        :rtype: String
        """
        logging.info("Crinkle submit()")
        if self.analyzer is None:
            raise Exception(f"Analyzer must be added before Crinkle slice submission using add_analyzer()")
        
        if (self.do_allocate_hosts): self.allocate_hosts()
        
        return super().submit(wait=wait, wait_timeout=wait_timeout, wait_interval=wait_interval, progress=progress, wait_jupyter=wait_jupyter, post_boot_config=post_boot_config, wait_ssh=wait_ssh,
                       extra_ssh_keys=extra_ssh_keys, lease_start_time=lease_start_time, lease_end_time=lease_end_time, lease_in_hours=lease_in_hours, validate=validate)
    
    def setup_ptp(self):
        prereq_cmd = f"sudo apt update && sudo apt install -y ansible git"
        git_cmd = f"cd {REMOTEWORKDIR} && git clone https://github.com/fabric-testbed/ptp.git"
        ansible_cmd = f"cd {REMOTEWORKDIR}/ptp/ansible && ansible-playbook --connection=local --inventory 127.0.0.1, --limit 127.0.0.1 playbook_fabric_experiment_ptp.yml"
        jobs = []
        site_ads = {}
        logging.info(f"Setting up PTP on Crinkle nodes")
        print("Setting up PTP on Crinkle nodes")
        jobs.append(self.analyzer.execute_thread(f"{prereq_cmd} && {git_cmd} && {ansible_cmd}"))
        for monitor_name, monitor in self.monitors.items():
            monitor_site = monitor.get_site()
            site_ad = site_ads.setdefault(monitor_site, self.get_fablib_manager().get_site_advertisement(monitor_site))
            if not site_ad.flags.ptp:
                logging.warning(f"Site {monitor_site} does not support PTP, skipping setup for monitor {monitor_name}")
                print(f"Site {monitor_site} does not support PTP, skipping setup for monitor {monitor_name}")
                continue
            jobs.append(monitor.execute_thread(f"{prereq_cmd} && {git_cmd} && {ansible_cmd}"))
        ctr = 0
        for _ in futures.as_completed(jobs):
            ctr += 1
            logging.info(f'{ctr}/{len(jobs)} jobs finished')
            print(f'{ctr}/{len(jobs)} jobs finished')
        self.do_ptp_setup = False

    def post_boot_config(self):
        """
        Runs post_boot_config identically to Slice.post_boot_config before running
        Crinkle specific functions. These are to:
            - Install the SPADE provenance analyzer to the analyzer node
            - Install other needed libraries and software to the analyzer and monitors
            - Update the slice local variables with new references to FABRIC resources
            - Create the monitor.data entries and save them to FABRIC
            - Start the SPADE instance and analyzer listening service

        Only use this method after a non-blocking submit call and only call it
        once.
        """
        # Rewritten super to handle Crinkle-specific nodes
        if self.is_dead_or_closing() or self.is_allocated():
            print(
                f"FAILURE: Slice is in {self.get_state()} state; cannot do post boot config"
            )
            return
        

        logging.info(
            f"post_boot_config: slice_name: {self.get_name()}, slice_id {self.get_slice_id()}"
        )

        # Make sure we have the latest topology
        self.update()

        for network in self.get_networks():
            network.config()

        for interface in self.get_all_interfaces():
            try:
                interface.config_vlan_iface()
            except Exception as e:
                logging.error(f"Interface: {interface.get_name()} failed to config")
                logging.error(e, exc_info=True)

        for interface in self.get_all_interfaces():
            try:
                interface.get_node().execute(
                    f"sudo nmcli device set {interface.get_device_name()} managed no",
                    quiet=True,
                )
            except Exception as e:
                logging.error(
                    f"Interface: {interface.get_name()} failed to become unmanaged"
                )
                logging.error(e, exc_info=True)

        import time

        start = time.time()

        my_thread_pool_executor = futures.ThreadPoolExecutor(32)
        threads = {}

        for node in self.get_all_nodes():
            # Run configuration on newly created nodes and on modify.
            logging.info(
                f"Configuring {node.get_name()} "
                f"(instantiated: {node.is_instantiated()}, "
                f"modify: {self._is_modify()})"
            )
            if not node.is_instantiated() or self._is_modify():
                thread = my_thread_pool_executor.submit(node.config)
                threads[thread] = node

        print("Running post boot config threads ...")

        for thread in futures.as_completed(threads.keys()):
            node = threads[thread]
            try:
                result = thread.result()
                print(
                    f"Post boot config {node.get_name()}, Done! ({time.time() - start:.0f} sec)"
                )
            except Exception as e:
                print(
                    f"Post boot config {node.get_name()}, Failed! ({time.time() - start:.0f} sec)"
                )
                logging.error(
                    f"Post boot config {node.get_name()}, Failed! ({time.time() - start:.0f} sec) {e}"
                )

        # Push updates to user_data
        print("Saving fablib data... ", end="")
        self.submit(wait=True, progress=False, post_boot_config=False, wait_ssh=False)
        self.update()

        for node in self.get_nodes():
            if "attestable_switch_config" in node.get_user_data():
                logging.info(
                    f"switch config: {str(node.get_user_data()['attestable_switch_config'])}"
                )
                aswitch = self.get_attestable_switch(name=node.get_name())
                aswitch.switch_config()

        # Custom Crinkle logic
        logging.info(f"Crinkle post_boot_config")
        if self.do_post_boot:
            self.analyzer = self.get_node(name=self.analyzer_name)
            site = self.analyzer.get_site()
            self.analyzer_cnet = self.get_l3network(name=f"{self.prefix}_ananet_{site}")
            self.analyzer_iface = self.analyzer.get_interface(network_name=self.analyzer_cnet.get_name())
            self.cnets[site]=self.analyzer_cnet
            jobs: list[futures.Future] = []
            counter = 0
            self.monitor_string = f"{self.analyzer_iface.get_device_name()} "
            for key, monitor in self.monitors.items():
                logging.info(f"Refreshing monitor for net {key} after slice creation, count {counter}")
                refreshed_monitor = self.get_monitor(monitor.get_name())
                mon_site = refreshed_monitor.get_site()
                jobs.append(refreshed_monitor.execute_thread(f"sudo ip link set {self.analyzer_iface.get_device_name()} up; wget -q -O {REMOTEWORKDIR}/{DPDKNAME} {MONITORURL}; chmod u+x {REMOTEWORKDIR}/{DPDKNAME}"))
                jobs.append(refreshed_monitor.execute_thread(f"sudo sed -i 's/^GRUB_CMDLINE_LINUX=.*/GRUB_CMDLINE_LINUX=\"default_hugepagesz=1G hugepagesz=1G hugepages={refreshed_monitor.get_ram()//2}\"/' /etc/default/grub && sudo grub-mkconfig -o /boot/grub/grub.cfg"))
                if self.cnets[mon_site] is None or not self.cnets[mon_site].is_instantiated():
                    self.cnets[mon_site] = self.get_l3network(name=f"{self.prefix}_ananet_{mon_site}")
                refreshed_monitor.data.cnet_iface = refreshed_monitor.get_interface(network_name=f"{self.prefix}_ananet_{mon_site}")
                ordered_devs = {}
                ctr = 0
                # Need a way to know where a specific interface falls *in the OS ordering of them*
                # Such as, if an interface is device enp8s0, and OS has enp7s0, enp8s0, enp9s0, return 1
                # And because this isn't consistent per site, cannot assume enp7s0 is first
                for entry in refreshed_monitor.get_dataplane_os_interfaces():
                    ordered_devs[entry['ifname']] = ctr
                    ctr += 1
                refreshed_monitor.data.cmd_args += f"-n {refreshed_monitor.data.monitor_id} "
                refreshed_monitor.data.cmd_args += f"-m {refreshed_monitor.data.cnet_iface.get_mac()} -m {self.analyzer_iface.get_mac()} "
                refreshed_monitor.data.cmd_args += f"-i {refreshed_monitor.data.cnet_iface.get_ip_addr()} -i {self.analyzer_iface.get_ip_addr()} "
                dev_name = refreshed_monitor.data.cnet_iface.get_device_name()
                if dev_name is None:
                    raise Exception(f"Monitor {refreshed_monitor.get_name()} failed to attach interfaces - please try another site")
                refreshed_monitor.data.cmd_args += f"-d {refreshed_monitor.data.port_nums}@{ordered_devs[dev_name]} "
                refreshed_monitor.data.port_nums += 1
                self.monitor_string += f'{refreshed_monitor.data.monitor_id} '
                for data in monitor.creation_data:
                    logging.info(f"Initializing monitor after slice creation with data: {data}")
                    mon_iface = refreshed_monitor.get_component(name=data[MonNetData.MIFACENAME]).get_interfaces()[0]
                    dev_name = mon_iface.get_device_name()
                    refreshed_monitor.data.cmd_args += f"-d {refreshed_monitor.data.port_nums}@{ordered_devs[dev_name]} "
                    refreshed_monitor.data.iface_mappings[data[MonNetData.IFACENAME]] = (
                        data[MonNetData.NODENAME], mon_iface, data[MonNetData.ISSINK], refreshed_monitor.data.port_nums)
                    self.monitor_string += f'{refreshed_monitor.data.port_nums}@{data[MonNetData.IFACENAME]} '
                    refreshed_monitor.data.port_nums += 1
                self.monitor_string += ' '
                self.monitors[key] = refreshed_monitor
                refreshed_monitor.set_monitor_data()
                counter += 1
            self.monitor_string = self.monitor_string.rstrip()
            logging.info(f"Crinkle post_boot_config waiting on jobs to finish")
            print("Configuring Crinkle Resources")
            ctr = 0
            for _ in futures.as_completed(jobs):
                ctr += 1
                logging.info(f'{ctr}/{len(jobs)} jobs finished')
                print(f'{ctr}/{len(jobs)} jobs finished')
            logging.info(f'Starting SPADE')
            self.analyzer.execute_thread(f'./{REMOTEWORKDIR}/SPADE/bin/spade debug')
            time.sleep(5)
            spade_control_commands = ('add reporter DSL /home/ubuntu/spade_pipe\n'
                                    'add analyzer CommandLine\n'
                                    'add storage Neo4j database=/home/ubuntu/spade_database\n')
            self.analyzer.execute(f'echo -e "{spade_control_commands}" | ./{REMOTEWORKDIR}/SPADE/bin/spade control', quiet=True)
            self.analyzer.upload_file(f"{CREASEDIR}/spade_reader.py", f"{REMOTEWORKDIR}/spade_reader.py")
            self.analyzer.execute(f"sudo chmod u+x {REMOTEWORKDIR}/spade_reader.py")
            if self.monitor_string == '':
                self.reset_monitor_string()
            self.analyzer.execute_thread(f"sudo ./{REMOTEWORKDIR}/spade_reader.py {self.monitor_string}")
            logging.info(f"Saving slice data before rebooting monitors")
            print("Saving slice data before rebooting monitors")
            self.do_post_boot = False
            if self.do_ptp_setup: self.setup_ptp()
            logging.info(f"Saving Crinkle Data")
            self.set_crinkle_data()
            self.submit(wait=True, progress=False, post_boot_config=False, wait_ssh=False)
            self.update()

            logging.info(f"Rebooting Crinkle monitors")
            print("Rebooting Crinkle Resources")
            for monitor in self.monitors.values():
                monitor.execute("sudo reboot")
            logging.info(f"Waiting on Crinkle monitors to finish reboot")
            self.wait_ssh(progress=True)
            jobs = []
            logging.info(f"Enabling Crinkle monitor interfaces")
            for monitor in self.monitors.values():
                cmd = f'sudo ip link set {monitor.data.cnet_iface.get_device_name()} up; sudo ip link set {monitor.data.cnet_iface.get_device_name()} promisc on; '
                for (_, iface, _, _) in monitor.data.iface_mappings.values():
                    dev_name = iface.get_device_name()
                    cmd += f'sudo ip link set {dev_name} up; sudo ip link set {dev_name} promisc on; '
                jobs.append(monitor.execute_thread(cmd.rstrip()))
            ctr = 0
            for _ in futures.as_completed(jobs):
                ctr += 1
                logging.info(f'{ctr}/{len(jobs)} jobs finished')
                print(f'{ctr}/{len(jobs)} jobs finished')
            logging.info(f"Crinkle post_boot_config done")
            print("Crinkle post_boot_config done")

    def get_nodes(self, refresh: bool = False) -> list[Node]:
        """
        Gets a list of all non-Crinkle nodes in this slice.

        :return: a list of fablib nodes
        :rtype: List[Node]
        """
        if not self.nodes or not len(self.nodes):
            refresh = True
        nodes = super().get_nodes(refresh=refresh)
        crinkle_nodes = [mon_name for mon_name in self.monitors.keys()]
        crinkle_nodes.append(self.analyzer_name)
        i = 0
        while i < len(nodes):
            node = nodes[i]
            nodename: str = node.get_name()
            if nodename.startswith(f'{self.prefix}_monitor_') or nodename == f'{self.prefix}_analyzer':
                nodes.pop(i)
            else:
                i += 1
        return nodes
    
    def get_all_nodes(self, refresh: bool = False):
        """
        Gets a list of all nodes in this slice.

        :return: a list of fablib nodes
        :rtype: List[Node]
        """
        return super().get_nodes(refresh=refresh)

    def get_interface(self, name: str = None, refresh: bool = False) -> Interface:
        """
        Gets a particular interface from this slice.

        :param name: the name of the interface to search for
        :type name: str

        :param refresh: Refresh the interface object with latest Fim info
        :type refresh: bool

        :raises Exception: if no interfaces with name are found
        :return: an interface on this slice
        :rtype: Interface
        """
        ret_val = self.get_all_interfaces(refresh=refresh, output="dict").get(name)
        if not ret_val:
            raise Exception("Interface not found: {}".format(name))
        return ret_val
    
    def get_interfaces(
        self, include_subs: bool = True, refresh: bool = False, output: str = "list"
        ) -> Union[dict[str, Interface], list[Interface]]:
        """
        Gets all non-Crinkle interfaces in this slice.

        :param include_subs: Flag indicating if sub interfaces should be included
        :type include_subs: bool

        :param refresh: Refresh the interface object with latest Fim info
        :type refresh: bool

        :param output: Specify how the return type is expected; Possible values: list or dict
        :type output: str

        :return: a list of interfaces on this slice
        :rtype: Union[dict[str, Interface], list[Interface]]
        """
        return super().get_interfaces(include_subs=include_subs, refresh=refresh, output=output)
    
    def get_all_interfaces(
        self, include_subs: bool = True, refresh: bool = False, output: str = "list"
        ) -> Union[dict[str, Interface], list[Interface]]:
        """
        Gets all interfaces in this slice.

        :param include_subs: Flag indicating if sub interfaces should be included
        :type include_subs: bool

        :param refresh: Refresh the interface object with latest Fim info
        :type refresh: bool

        :param output: Specify how the return type is expected; Possible values: list or dict
        :type output: str

        :return: a list of interfaces on this slice
        :rtype: Union[dict[str, Interface], list[Interface]]
        """
        if len(self.all_interfaces) == 0 or refresh:
            for node in self.get_all_nodes(refresh=refresh):
                logging.debug(f"Getting interfaces for node {node.get_name()}")
                # get_nodes will already refresh interfaces if needed
                n_ifaces = node.get_interfaces(include_subs=include_subs, output="dict")
                self.all_interfaces.update(n_ifaces)
            for fac in self.get_facilities(refresh=refresh):
                logging.debug(f"Getting interfaces for facility {fac.get_name()}")
                fac_ifaces = fac.get_interfaces(refresh=refresh, output="dict")
                self.all_interfaces.update(fac_ifaces)

        if output == "dict":
            return self.all_interfaces
        else:
            return list(self.all_interfaces.values())
    
    @staticmethod
    def mac_to_int(mac: str):
        """
        Translates a MAC address to an int

        :param mac: MAC address
        :type mac: String
        :return: The integer value of the MAC address
        :rtype: int
        """
        hexes = mac.split(':')
        intval = 0
        for i in range(0, 6):
            hexval = hexes[5-i]
            intval += (16**(2*i+1))*int(hexval[0], 16) + (16**(2*i))*int(hexval[1], 16)
        return intval
    
    @staticmethod
    def ip6_to_int(ip6: str):
        """
        Translates an IPv6 address to an int

        :param mac: IPv6 address
        :type mac: String
        :return: A tuple of the integer values of the front and back halves
        :rtype: tuple[int, int]
        """
        ip6 = IPv6Address(ip6)
        return (int(ip6)//(2**64),int(ip6)%(2**64))
    
    def trace_ping(self, src_ip: str, dst_ip: str, iface_name: str = "", iface: Interface = None, name: str = "trace_ping"):
        """
        Create a ping packet from src_ip to dst_ip, that will be sent into the specified
        monitored interface. This will return a graph of the history of the request as it traversed
        the network.
        """
        self.probe(f'Ether(dst="ff:ff:ff:ff:ff:ff")/IP(src="{src_ip}",dst="{dst_ip}")/ICMP()',
                   iface_name=iface_name, iface=iface, name=name)
    
    def probe(self, scapy: str, iface_name: str = "", iface: Interface = None, name: str = "probe"):
        """
        Send a packet, formed from the given scapy definition and appended with a unique id,
        into the targeted interface then download the graph of its traversal across
        the network.

        :param scapy: Scapy definition the packet will be crafted from, as a string
        :type scapy: String
        :param iface_name: The name of the interface to target with the packet
        :type iface_name: String
        :param iface: An interface to target with the packet
        :param name: The name of the downloaded graph
        :type name: String
        """
        if iface_name == "" and iface is not None:
            iface_name = iface.get_name()
        elif iface_name == "" and iface is None:
            raise Exception("Exception: Either iface_name or iface must not be empty")
        
        net = self.get_interface(name=iface_name).get_network()
        net_name = net.get_name()
        if not net_name.startswith(f'{self.prefix}_net_'):
            raise Exception("Exception: interface must be part of a monitored network")
        
        orig_net_name = self.get_orig_net_name(net)
        monitor = None
        if orig_net_name == "":
            ifaces = net.get_interfaces()
            for net_iface in ifaces:
                node_name = net_iface.get_node().get_name()
                if node_name.startswith(f"{self.prefix}_monitor_"):
                    monitor = self.get_monitor(node_name)
        else:
            monitor = self.monitors[orig_net_name]
        
        port = monitor.data.iface_mappings[iface_name][3]
        dev_name = monitor.data.iface_mappings[iface_name][1].get_device_name()
        if port == 1:
            port = 2
        else:
            port = 1
        scapy = scapy.replace('"', '\\"')
        scapy = scapy.replace("'", "\\'")
        # uid_trailer[1] = (mon_id << 48) + (port << 32) + ((uid_trailer[0] << 16) & 0x00000000FFFF0000) + MONPROT;
        uid = (self.probe_id << 64) + (monitor.data.monitor_id << 48) + (port << 32) + ((self.probe_id << 16) & 0x00000000FFFF0000) + MONPROT
        new_scapy = f'Raw({scapy})/Raw(int({uid}).to_bytes(16, \\"big\\"))'
        monitor.execute(f'''echo -e "from scapy.all import *\npkt={new_scapy}\nsendp(pkt, iface='{dev_name}')\n" | sudo python3''')
        logging.info(f'Sent probe packet with uid {monitor.data.monitor_id}-{port}-{self.probe_id}')
        self.get_graph(name=name, pkt_id=f'{monitor.data.monitor_id}-{port}-{self.probe_id}')
        self.probe_id += 1

    def dump_counters(self) -> dict[str, dict[str, tuple[str, int, int]]]:
        """
        Runs through each non-crinkle Node, gets the rx_packets and tx_packets count for
        each interface, then returns a dict of form dict[node name, dict[iface name, (dev name, rx, tx)]].
        """
        rdict: dict[str, dict[str, tuple[str, int, int]]] = {}
        for node in self.get_non_crinkle_nodes():
            if node.get_image() in UBUNTU_IMAGES:
                nodename = node.get_name()
                for iface in node.get_interfaces():
                    ifacename = iface.get_name()
                    devname = iface.get_device_name()
                    stdout, _ = node.execute(f"cat /sys/class/net/{devname}/statistics/rx_packets && cat /sys/class/net/{devname}/statistics/tx_packets", quiet=True)
                    counters = stdout.splitlines()
                    rx = int(counters[0])
                    tx = int(counters[1])
                    if nodename in rdict:
                        rdict[nodename][ifacename] = (devname, rx, tx)
                    else:
                        rdict[nodename] = {ifacename: (devname, rx, tx)}
        return rdict

    def list_counters(self,
                      output=None,
                      fields=None,
                      filter_function=None,
                      quiet=False,
                      pretty_names=True):
        table = []
        counter_dict = self.dump_counters()

        pretty_names_dict = {}
        if pretty_names:
            pretty_names_dict = {
                "node_name": "Node Name",
                "interface": "Interface",
                "dev_name": "Device Name",
                "rx_pkts": "RX Packets",
                "tx_pkts": "TX Packets"
            }

        for nodename, nodedict in counter_dict.items():
            for ifacename, ifacetuple in nodedict.items():
                rowdict = {"node_name": nodename,
                           "interface": ifacename,
                           "dev_name": ifacetuple[0],
                           "rx_pkts": ifacetuple[1],
                           "tx_pkts": ifacetuple[2]}
                table.append(rowdict)

        table = sorted(table, key=lambda x: (x["node_name"], x["interface"]))

        table = self.get_fablib_manager().list_table(
            table,
            fields=fields,
            title="Interface Counters",
            output=output,
            quiet=True,
            filter_function=filter_function,
            pretty_names_dict=pretty_names_dict
        )

        if table and not quiet:
            display(table)

        return table


    @staticmethod
    def ip_net_like(net: str):
        """
        Not intended for API use. Transforms IPv4 subnets into LIKE terms for SPADE.

        :param net: The IPv4 subnet, either of the form X.0.0.0/8, X.Y.0.0/16, X.Y.Z.0/24,
            or X, X.Y, X.Y.Z
        :type net: String
        :return: the LIKE term for SPADE
        :rtype: String
        """
        net_mask = net.split('/')
        ret_str = ""
        net_parts = net_mask[0].split('.')
        ctr = 0
        if net_parts[ctr] == "0":
            ret_str += '%'
        else:
            ret_str += net_parts[ctr]
        ctr += 1
        while ctr < len(net_parts):
            if net_parts[ctr] == "0":
                ret_str += '.%'
            else:
                ret_str += '.'+net_parts[ctr]
            ctr += 1
        if ctr < 3:
            ret_str += '.%'
        return ret_str
                

    def get_graph(self, name: str = "graph", filterin: str = None, tstart = None, tend = None, tformat="epoch", pkt_id = None, quiet=True):
        """
        Produce and download a graph of the data stored in the analyzer's SPADE database,
        filtered using the given arguments.

        :param name: The name of the downloaded graph, default "graph".
        :type name: String
        :param filterin: A tcpdump filter (currently supports ip, tcp, udp, icmp) to filter flows in the graph on
        :type filterin: String
        :param tstart: A time filter, such that the returned graph is for all packets with time >= tstart. Default off
        :type tstart: int
        :param tend: A time filter, such that the returned graph is for all packets with time <= tend. Default off
        :type tend: int
        :param pkt_id: Only return packets matching the id
        :type pkt_id: int
        :param quiet: If True, suppresses SPADE query output and just returns the graph
        :type quiet: bool
        """
        spade_filter = ""
        if filterin is not None:
            filter_words = filterin.split(' ')
            i = 0
            ctr = 0
            while i < len(filter_words) and ctr < 100:
                do_ip = False
                do_port = False
                if filter_words[i] == 'and' or filter_words[i] == 'or':
                    spade_filter += f'{filter_words[i]} '
                    i += 1
                if filter_words[i] == 'src':
                    spade_filter += '\\"ip.src\\"'
                    do_ip = True
                    i += 1
                elif filter_words[i] == 'dst':
                    spade_filter += '\\"ip.dst\\"'
                    do_ip = True
                    i += 1
                elif filter_words[i] == 'ip':
                    spade_filter += '''\\"eth.type\\" == '0x800' '''
                    if i + 1 < len(filter_words):
                        do_ip = True
                    i += 1
                elif filter_words[i] == 'tcp':
                    spade_filter += '''\\"ip.prot\\" == '6' '''
                    do_port = True
                    i += 1
                elif filter_words[i] == 'udp':
                    spade_filter += '''\\"ip.prot\\" == '17' '''
                    do_port = True
                    i += 1
                elif filter_words[i] == 'icmp':
                    spade_filter += '''\\"ip.prot\\" == '1' '''
                    i += 1
                elif filter_words[i] == 'host':
                    spade_filter += f'''\\"ip.src\\" == '{filter_words[i+1]}' or ip.dst == '{filter_words[i+1]}' '''
                    i += 2
                elif filter_words[i] == 'net':
                    spade_filter += f'''\\"ip.src\\" LIKE '{self.ip_net_like(filter_words[i+1])}' or ip.dst LIKE '{self.ip_net_like(filter_words[i+1])}' '''
                    i += 2
                if do_ip:
                    if filter_words[i] == 'src':
                        spade_filter += 'and \\"ip.src\\"'
                        i += 1
                    elif filter_words[i] == 'dst':
                        spade_filter += 'and \\"ip.dst\\"'
                        i += 1
                    if filter_words[i] == 'host':
                        if filter_words[i-1] == 'src' or filter_words[i-1] == 'dst':
                            spade_filter += f''' == '{filter_words[i+1]}' '''
                        else:
                            spade_filter += f'''and \\"ip.src\\" == '{filter_words[i+1]}' or ip.dst == '{filter_words[i+1]}' '''
                        i += 2
                    elif filter_words[i] == 'net':
                        if filter_words[i-1] == 'src' or filter_words[i-1] == 'dst':
                            spade_filter += f'''\\" LIKE '{self.ip_net_like(filter_words[i+1])}' '''
                        else:
                            spade_filter += f'''and \\"ip.src\\" LIKE '{self.ip_net_like(filter_words[i+1])}' or ip.dst LIKE '{self.ip_net_like(filter_words[i+1])}' '''
                        i += 2
                    else:
                        raise Exception(f"Invalid term following {filter_words[i]}: {filter_words[i+1]}")
                if do_port:
                    if i >= len(filter_words): break
                    if filter_words[i] == 'port':
                        spade_filter += f'''and \\"prot.sport\\" == '{filter_words[i+1]}' or \\"prot.dport\\" == '{filter_words[i+1]}' '''
                        i += 2
                    elif filter_words[i] == 'src':
                        spade_filter += f'''and \\"prot.sport\\"  == '{filter_words[i+1]}' '''
                        i += 2
                    elif filter_words[i] == 'dst':
                        spade_filter += f'''and \\"prot.dport\\"  == '{filter_words[i+1]}' '''
                        i += 2
                ctr += 1
        spade_filter = spade_filter.rstrip()
        time_filter = ""
        if tformat != "epoch":
            if tstart:
                tstart: datetime = datetime.strptime(tstart, tformat)
                tstart = tstart.timestamp() * 1000000
            if tend:
                tend: datetime = datetime.strptime(tend, tformat)
                tend = tend.timestamp() * 1000000
        if tstart and tend:
            time_filter += f'''(\\"epoch\\" >= '{tstart}' and \\"epoch\\" <= '{tend}')'''
        elif tstart:
            time_filter += f'''(\\"epoch\\" >= '{tstart}')'''
        elif tend:
            time_filter += f'''(\\"epoch\\" <= '{tend}')'''
        graph_build = ""
        if pkt_id:
            graph_build += f'''\\$graph0 = \\$base.getPath(\\$base.getEdge(\\"pkt_id\\" == '{pkt_id}').limit(1000).getEdgeEndpoints(), \\$base.getVertex(\\"type\\" == 'Agent'), 1) + \\$base.getEdge(\\"pkt_id\\" == '{pkt_id}').limit(1000) + \\$base.getEdge(\\"pkt_id\\" == '{pkt_id}').limit(1000).getEdgeEndpoints()\n'''
        else:
            graph_build += '''\\$graph0 = \\$base\n'''
        if time_filter != "":
            graph_build += f'''\\$graph1 = \\$graph0.getPath(\\$graph0.getEdge({time_filter}).limit(1000).getEdgeEndpoints(), \\$graph0.getVertex(\\"type\\" == 'Agent'), 1) + \\$graph0.getEdge({time_filter}).limit(1000) + \\$graph0.getEdge({time_filter}).limit(1000).getEdgeEndpoints()\n'''
        else:
            graph_build += '''\\$graph1 = \\$graph0\n'''
        graph_build += f'''\\$graph2 = \\$graph1.getLineage(\\$graph1.getVertex({spade_filter}), 1, 'b')\n\\$graph3 = \\$graph2 + \\$base.getPath(\\$graph2.getVertex(\\"type\\" == 'Process'), \\$base.getVertex(\\"type\\" == 'Agent'), 1)'''
        self.analyzer.execute(f'echo -e "set storage Neo4j\n{graph_build}\nexport > /home/ubuntu/{REMOTEWORKDIR}/{name}.dot\ndump all \\$graph3" | ./{REMOTEWORKDIR}/SPADE/bin/spade query; '
                                f'dot -Tsvg {REMOTEWORKDIR}/{name}.dot -o {REMOTEWORKDIR}/{name}.svg', quiet=quiet)
        self.analyzer.download_file(f'{name}.svg', f'{REMOTEWORKDIR}/{name}.svg')

    def reset_monitor_string(self):
        logging.info("Crinkle tried to start with blank monitor string, regenerating it")
        self.monitor_string = f"{self.analyzer_iface.get_device_name()} "
        for monitor in self.monitors.values():
            self.monitor_string += f'{monitor.data.monitor_id} '
            for iface_name, (_, _, _, num) in monitor.data.iface_mappings.items():
                self.monitor_string += f'{num}@{iface_name} '
        self.monitor_string = self.monitor_string.rstrip()

    def reset_analyzer(self, quiet: bool = True):
        """
        Reset the analyzer, wiping the database.
        """
        if self.monitor_string == '':
            self.reset_monitor_string()
        self.analyzer.execute(f'''sudo killall python3; echo -e "remove storage Neo4j\n" | ./{REMOTEWORKDIR}/SPADE/bin/spade control; rm -rf {REMOTEWORKDIR}/spade_database/; echo -e "add storage Neo4j database=/home/ubuntu/{REMOTEWORKDIR}/spade_database\n" | ./{REMOTEWORKDIR}/SPADE/bin/spade control''', quiet=quiet)
        self.analyzer.execute_thread(f'sudo ./{REMOTEWORKDIR}/spade_reader.py {self.monitor_string}')
    
    def start_monitor(self, monitor: CrinkleMonitor, wait: bool=True, quiet: bool=False) -> futures.Future | None:
        """
        Start the DPDK script on a monitor.

        :param monitor: The monitor node to target
        :type monitor: CrinkleMonitor
        :param wait: Whether this call is blocking, default True
        :type wait: bool
        :param quiet: Whether to print stdout/stderr of a blocking call, default False
        :type quiet: bool
        :return: The future of a non-blocking call, or None
        :rtype: concurrent.futures.Future
        """
        if monitor is None:
            raise Exception("Monitor cannot be None")
        logging.info(f"Starting Crinkle monitor {monitor.data.net_name}")
        job = None
        if wait:
            monitor.execute(f"sudo ./{REMOTEWORKDIR}/{DPDKNAME} -- {monitor.data.cmd_args} &", quiet=quiet)
        else:
            job = monitor.execute_thread(f"sudo ./{REMOTEWORKDIR}/{DPDKNAME} -- {monitor.data.cmd_args}")
        return job
        
    def start_all_monitors(self, wait: bool=True, no_return: bool=True) -> list[futures.Future] | None:
        """
        Start the DPDK script on all monitors.

        :param wait: Whether this call is blocking, default True
        :type wait: bool
        :param no_return: Whether to return None instead of the list of jobs
        :type no_return: bool
        :return: The futures of all started jobs
        :rtype: list[concurrent.futures.Future]
        """
        start_list: list[futures.Future] = []
        for monitor in self.monitors.values():
            start_list.append(self.start_monitor(monitor=monitor, wait=False))
        if wait:
            logging.info(f"Waiting for monitors to finish starting")
            ctr = 0
            max = len(start_list)
            while ctr < max:
                if start_list[ctr].running():
                    ctr += 1
                    logging.info(f'{ctr}/{max} started')
        if (no_return):
            return
        return start_list

    def stop_monitor(self, monitor: CrinkleMonitor):
        if monitor is None:
            raise Exception("monitor cannot be None")
        monitor.execute(f"sudo killall {DPDKNAME}")

    def stop_all_monitors(self):
        for monitor in self.monitors.values():
            self.stop_monitor(monitor=monitor)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
from recordclass import recordclass

from .base_sliver import BaseSliver
from .network_service import NSLayer
from .topology_diff import TopologyDiff


class LinkType(enum.Enum):
    """
    Possible Link types in FABRIC. Links are passive, cannot be instantiated
    on demand. Links can only connect two interfaces.
    """
    Patch = enum.auto() # DAC cable or patch fiber
    L1Path = enum.auto() # Wave
    L2Path = enum.auto() # provider L2 path

    def help(self) -> str:
        return self.name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        for lt in LinkType:
            if s == lt.name:
                return cls(lt)
        return None


LinkConstraintRecord = recordclass('LinkConstraintRecord',
                                   ['layer', 'desc',
                                    'num_interfaces'])


class NetworkLinkSliver(BaseSliver):

    NAME_REGEX = r'^[\w\-+_/\.\ :]{2,255}$'

    """
    Services can be limited by the number of interfaces/connection points they can connect
    The number of sites they may connect and the number of instances they may have in a slice.
    """
    # whenever there is no limit, num is set to 0
    NO_LIMIT = 0

    LinkConstraints = {
        LinkType.Patch: LinkConstraintRecord(layer=NSLayer.L2, num_interfaces=2,
                                             desc='A fiber patch or a cable.'),
        LinkType.L1Path: LinkConstraintRecord(layer=NSLayer.L1, num_interfaces=2,
                                             desc='A wavelength.'),
        LinkType.L2Path: LinkConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT,
                                              desc='A provider L2 path or service.')}

    def __init__(self):
        super().__init__()
        self.layer = None
        self.technology = None

    #
    # Setters are only needed for things we want users to be able to set
    #
    def get_layer(self) -> NSLayer:
        return self.layer

    def set_layer(self, layer: NSLayer):
        self.layer = layer

    def get_technology(self) -> str:
        return self.technology

    def set_technology(self, technology: str):
        self.technology = technology

    @staticmethod
    def type_from_str(ltype: str) -> LinkType or None:
        if ltype is None:
            return None
        return LinkType.from_string(ltype)

    @staticmethod
    def layer_from_str(layer: str) -> NSLayer or None:
        if layer is None:
            return None
        return NSLayer.from_string(layer)

    def diff(self, other_sliver) -> TopologyDiff or None:
        raise RuntimeError('Not implemented')



from lark import Lark, ParseTree, exceptions

from pathlib import Path

from rfc3987_syntax.utils import load_grammar

RFC3987_SYNTAX_PARSER_TYPE: str = "earley"
RFC3987_SYNTAX_GRAMMAR_PATH: Path = Path(__file__).parent / "syntax_rfc3987.lark"
RFC3987_SYNTAX_TERMS: list[str] = [
    "iri",
    "iri_reference",
    "absolute_iri",
    "scheme",
    "irelative_ref",
    "irelative_part"
    "ihier_part",
    "iauthority",
    "iuserinfo",
    "ihost",
    "ireg_name",
    "ipath_abempty",
    "isegment",
    "isegment_nz",
    "isegment_nz_nc",
    "ipchar",
    "iquery",
    "ifragment",
    "iunreserved",
    "ucschar",
    "iprivate",
    "sub_delims",
    "ip_literal",
    "ipvfuture",
    "ipv6address",
    "h16",
    "ls32",
    "ipv4address",
    "dec_octet",
    "digit",
    "non_zero",
    "unreserved",
    "alpha",
    "hexdig",
    "port",
    "pct_encoded",
]

grammar: str = load_grammar(RFC3987_SYNTAX_GRAMMAR_PATH)

syntax_parser = Lark(grammar, start=["iri", "iri_reference", "absolute_iri"], parser=RFC3987_SYNTAX_PARSER_TYPE)


def parse(term: str, value: str) -> ParseTree:
    return syntax_parser.parse(value, start=term)


def is_valid_syntax(term: str, value: str):
    try:
        parse(term=term, value=value)
        return True
    except exceptions.LarkError:
        return False


def make_syntax_validator(rule_name):
    parser = Lark(grammar, start=rule_name, parser=RFC3987_SYNTAX_PARSER_TYPE)

    def syntax_validator(text):
        try:
            parser.parse(text)
            return True
        except exceptions.LarkError:
            return False

    return syntax_validator


is_valid_syntax_iri = make_syntax_validator("iri")

is_valid_syntax_iri_reference = make_syntax_validator("iri_reference")

is_valid_syntax_absolute_iri = make_syntax_validator("absolute_iri")

is_valid_syntax_irelative_ref = make_syntax_validator("irelative_ref")

is_valid_syntax_irelative_part = make_syntax_validator("irelative_part")

is_valid_syntax_ihier_part = make_syntax_validator("ihier_part")

is_valid_syntax_iauthority = make_syntax_validator("iauthority")

is_valid_syntax_iuserinfo = make_syntax_validator("iuserinfo")

is_valid_syntax_ihost = make_syntax_validator("ihost")

is_valid_syntax_ireg_name = make_syntax_validator("ireg_name")

is_valid_syntax_ipath = make_syntax_validator("ipath")

is_valid_syntax_ipath_abempty = make_syntax_validator("ipath_abempty")

is_valid_syntax_ipath_absolute = make_syntax_validator("ipath_absolute")

is_valid_syntax_ipath_noscheme = make_syntax_validator("ipath_noscheme")

is_valid_syntax_ipath_rootless = make_syntax_validator("ipath_rootless")

is_valid_syntax_ipath_empty = make_syntax_validator("ipath_empty")

is_valid_syntax_isegment = make_syntax_validator("isegment")

is_valid_syntax_isegment_nz = make_syntax_validator("isegment_nz")

is_valid_syntax_isegment_nz_nc = make_syntax_validator("isegment_nz_nc")

is_valid_syntax_ipchar = make_syntax_validator("ipchar")

is_valid_syntax_iquery = make_syntax_validator("iquery")

is_valid_syntax_ifragment = make_syntax_validator("ifragment")

is_valid_syntax_iunreserved = make_syntax_validator("iunreserved")

is_valid_syntax_ucschar = make_syntax_validator("ucschar")

is_valid_syntax_iprivate = make_syntax_validator("iprivate")

is_valid_syntax_sub_delims = make_syntax_validator("sub_delims")

is_valid_syntax_ip_literal = make_syntax_validator("ip_literal")

is_valid_syntax_ipvfuture = make_syntax_validator("ipvfuture")

is_valid_syntax_ipv6address = make_syntax_validator("ipv6address")

is_valid_syntax_h16 = make_syntax_validator("h16")

is_valid_syntax_ls32 = make_syntax_validator("ls32")

is_valid_syntax_ipv4address = make_syntax_validator("ipv4address")

is_valid_syntax_dec_octet = make_syntax_validator("dec_octet")

is_valid_syntax_unreserved = make_syntax_validator("unreserved")

is_valid_syntax_alpha = make_syntax_validator("alpha")

is_valid_syntax_digit = make_syntax_validator("digit")

is_valid_syntax_hexdig = make_syntax_validator("hexdig")

is_valid_syntax_port = make_syntax_validator("port")



import sys
import time
import json
import argparse
import logging
import ipaddress

from provp4_api import (
    record_activity,
    record_API_call,
    record_agent,
    close_channel,
)
from bmv2_controller import BMV2Controller

# Static global variables for index
REG_NAME_INDEX = 0
ACTION_NAME_INDEX = 1
VALUE_ANNOTATION_INDEX = 2
INDEX_SIZE_INDEX = 3
VALUE_SIZE_INDEX = 4

# CONST SIZES # TODO - Make this come from a config file
ACTIVITY_SIZE = 8  # 8 bits for activity ID - Should be computed
# ACTIVITY_SIZE = 3
OPERATION_SIZE = 1  # 1 bit for READ/WRITE tag
VERSION_SIZE = 8
# VERSION_SIZE = 3

REGISTER_SIZE = 65536
# PKT_IN_TYPE = {"ipid": "hex", "src_ip": "ip_addr"}
PKT_IN_TYPE = {"src_mac": "mac_addr", "packet_id": "hex", "ether_type": "hex"}
index_map = {"MyIngress.prov_MyIngress_status_register": 0}

logger = logging.getLogger("dp-prov")


class ProvCollector:
    """This class collects the provenance from the data plane and sends it to the ProvManager for processing"""

    def __init__(self, reg_prov_details, thrift_port, switch_name) -> None:
        self.reg_prov_details = reg_prov_details
        self.dp_ip = "localhost"
        self.dp_port = thrift_port
        self.cp_client = None
        self.controller = BMV2Controller("localhost", thrift_port)
        self.agent_id = self.controller.get_switch_info()
        self.switch_name = switch_name
        self.last_count_index = {"packet_in": 0, "packet_out": 0}
        self.last_count_table_rule = 0
        for prov_reg_name in self.reg_prov_details:
            self.last_count_index[prov_reg_name] = 0

    def _parse_prov_test(self, reg_name, total_size, entry_string) -> dict:
        total_bit_width = sum([x[1] for x in total_size])
        logger.info(f"total bit width: {total_bit_width}")
        prov_string = format(entry_string, "b").zfill(total_bit_width + OPERATION_SIZE)
        logger.info(f"reg name: {reg_name} and prov_string: {prov_string}")
        start_index = 0
        entity_value_dict = {}
        operation_tag = prov_string[-1]
        prov_string = prov_string[:-1]  # Remove the operation bit
        new_size = []
        entity_value_type = self.reg_prov_details.get(reg_name).get("value_type")
        if operation_tag == "0":
            # READ operation. Therfore, swap the index and value sizes
            # total_size[0], total_size[1] = total_size[1], total_size[0]
            new_size.append(total_size[1])
            new_size.append(total_size[0])
            new_size.extend(total_size[2:])
            # logger.info(f"size: {new_size}")
            entity_value_dict["operation"] = "READ"
        else:
            new_size.append(total_size[0])
            new_size.append(total_size[1])
            new_size.extend(total_size[2:])
            # logger.info(f"size: {new_size}")
            entity_value_dict["operation"] = "WRITE"
        for entity_field in new_size:
            entity_type = entity_field[0]
            entity_size = entity_field[1]
            end_index = start_index + entity_size
            entity_value = prov_string[start_index:end_index]
            if entity_type == "value" and entity_value_type == "ip_addr":
                entity_value_converted = ".".join(
                    [
                        str((int(entity_value, 2) >> (8 * i)) & 255)
                        for i in range(3, -1, -1)
                    ]
                )
            elif entity_type == "value" and entity_value_type == "mac_addr":
                entity_value_converted = "{:012x}".format(int(entity_value, 2))
                entity_value_converted = ":".join(
                    [entity_value_converted[i : i + 2] for i in range(0, 12, 2)]
                )
            else:
                entity_value_converted = int(entity_value, 2)
            start_index = end_index
            entity_value_dict.update({entity_type: entity_value_converted})
            # logger.info(f"{entity_type}, {entity_value_converted}")
        # if entity_value_dict["operation"] == 0:
        #     logger.info(f"Value: {entity_value_dict['value']} - READ")
        # elif entity_value_dict["operation"] == 1:
        #     logger.info(f"Value: {entity_value_dict['value']} - WRITE")
        # else:
        #     logger.info("Other")
        logger.info(entity_value_dict)
        # Add the activity name instead of the activity id
        activity_id = str(entity_value_dict["activity"])
        if activity_id == "255":
            print("[WARN] 255 detected")
        else:
            entity_value_dict.update(
                {"activity": self.reg_prov_details[reg_name][activity_id]}
            )
        return entity_value_dict

    def _prov_wraparound(self):
        entries_read = 0
        for reg_name in self.reg_prov_details:
            if (
                reg_name == "packet_in"
                or (reg_name == "table_rule")
                or (reg_name == "packet_out")
            ):
                continue
            # TODO - make this dynamic, carefully check the P4 variable names
            prov_reg_name = f"MyIngress.prov_{reg_name}"
            # prov_reg_name = f"ingress.firewall_control.prov_{reg_name}"
            # prov_reg_name = f"prov_{reg_name}"
            # prov_reg_name = f"MyIngress.prov_{reg_name}"
            try:
                regs = self.controller.read_register(prov_reg_name)
            except Exception as e:
                print(
                    f"[ERROR] Unable to read prov register (name - {prov_reg_name})...exiting"
                )
                logger.error(
                    f"Unable to read prov register (name - {prov_reg_name})...exiting",
                    e,
                )
                sys.exit(1)
            # Query the index array and save the last written index
            # prov_reg_indices = self.controller.read_register(
            #     "ingress.firewall_control.prov_index_reg"
            # )
            prov_reg_indices = self.controller.read_register("MyIngress.prov_index_reg")
            # prov_reg_indices = self.controller.read_register("prov_index_reg")
            end_idx = prov_reg_indices[index_map[prov_reg_name]]
            start_idx = self.last_count_index[reg_name]
            idx = start_idx

            record_agent(self.switch_name, self.agent_id, "Switch", time.time())
            prov_width_list = [
                ["index", self.reg_prov_details[reg_name]["index_size"]],
                ["value", self.reg_prov_details[reg_name]["value_size"]],
                ["activity", ACTIVITY_SIZE],
                ["version", VERSION_SIZE],
            ]

            while idx != end_idx:
                prov_entry = regs[idx]
                if prov_entry == 0:
                    print(f"Entry 0: idx: {idx} and end_idx:{end_idx}")
                    break
                # Parse the provenance record
                parsed_prov_values = self._parse_prov_test(
                    reg_name, prov_width_list, prov_entry
                )
                record_activity(
                    parsed_prov_values["activity"],
                    self.agent_id,
                    parsed_prov_values["version"],
                    "PacketProcessor",
                )
                converted_value = parsed_prov_values["value"]
                record_API_call(
                    reg_name,
                    self.agent_id,
                    parsed_prov_values["version"],
                    "register",
                    parsed_prov_values["operation"],
                    {
                        "index": parsed_prov_values["index"],
                        "value": converted_value,
                        "activity_name": parsed_prov_values["activity"],
                    },
                )
                logger.info(f"Register ({reg_name}) entity value: {parsed_prov_values}")
                entries_read += 1
                idx = (idx + 1) % REGISTER_SIZE
                # print(f"Entry: {prov_entry} and Index: {idx}")
            self.last_count_index[reg_name] = end_idx
            print(
                f"[INFO] {reg_name}: start={start_idx}, end={idx}, total_entries_read={entries_read}, {round((self.last_count_index[reg_name] / REGISTER_SIZE) * 100, 3)}"
            )
            # print(
            #     f"[INFO] Pulling Register ({reg_name}) provenance.. {round((self.last_count_index[reg_name] / REGISTER_SIZE) * 100, 3)}"
            # )

    def _read_prov_regs(self):
        for reg_name in self.reg_prov_details:
            if (
                reg_name == "packet_in"
                or (reg_name == "table_rule")
                or (reg_name == "packet_out")
            ):
                continue
            # TODO - make this dynamic, carefully check the P4 variable names
            # prov_reg_name = f"prov_{reg_name}"
            # prov_reg_name = f"ingress.firewall_control.prov_{reg_name}"
            prov_reg_name = f"MyIngress.prov_{reg_name}"
            try:
                regs = self.controller.read_register(prov_reg_name)
            except Exception as e:
                print(
                    f"[ERROR] Unable to read prov register (name - {prov_reg_name})...exiting"
                )
                logger.error(
                    f"Unable to read prov register (name - {prov_reg_name})...exiting",
                    e,
                )
                sys.exit(1)
            count = 0
            # register_name = self.reg_prov_details[prov_reg_name][REG_NAME_INDEX]
            # get the activity name after parsing prov record
            # activity_name = self.reg_prov_details[prov_reg_name][ACTION_NAME_INDEX]
            # value_annotation = self.reg_prov_details[prov_reg_name][
            #     VALUE_ANNOTATION_INDEX
            # ]
            # activity_sent = False
            # Send the agent and activity once
            record_agent(self.switch_name, self.agent_id, "Switch", time.time())
            prov_width_list = [
                ["index", self.reg_prov_details[reg_name]["index_size"]],
                ["value", self.reg_prov_details[reg_name]["value_size"]],
                ["activity", ACTIVITY_SIZE],
                ["version", VERSION_SIZE],
            ]
            for prov_entry in regs[self.last_count_index[reg_name] :]:
                if prov_entry != 0:
                    count += 1
                    # Parse the provenance record
                    parsed_prov_values = self._parse_prov_test(
                        reg_name, prov_width_list, prov_entry
                    )
                    # if not activity_sent:
                    record_activity(
                        parsed_prov_values["activity"],
                        self.agent_id,
                        parsed_prov_values["version"],
                        "PacketProcessor",
                    )
                    # activity_sent = True
                    # Convert the value if annotation is present
                    converted_value = parsed_prov_values["value"]
                    # if value_annotation:
                    #     converted_value = self._convert_value(
                    #         value_annotation, parsed_prov_values["value"]
                    #     )
                    # Send the provenance event to the message queue
                    record_API_call(
                        reg_name,
                        self.agent_id,
                        parsed_prov_values["version"],
                        "register",
                        parsed_prov_values["operation"],
                        {
                            "index": parsed_prov_values["index"],
                            "value": converted_value,
                            "activity_name": parsed_prov_values["activity"],
                        },
                    )
                    logger.info(
                        f"Register ({reg_name}) entity value: {parsed_prov_values}"
                    )
                else:
                    break
            self.last_count_index[reg_name] += count
            print(
                f"[INFO] Pulling Register ({reg_name}) provenance.. {round(self.last_count_index[reg_name] / 65535, 3) * 100}"
            )
            logger.info(
                f"[INFO] Provenance register capacity: {self.last_count_index[reg_name] / 65535}"
            )
    def _read_packet_in_prov(self):
        # TODO - Modify such that pkt_in details are taken from the details json file.
        # prov_width_list = [
        #     ["ipid", 16],
        #     ["src_ip", 32],
        #     ["activity", 7],
        #     ["version", 8],
        # ]

        # prov_width_list = [
        #     ["src_mac", 16],
        #     ["packet_id", 16],
        #     ["ingress_port", 8],
        #     ["ether_type", 16],
        #     ["version", 3],
        # ]
        prov_width_list = [
            ["packet_id", 16],
            ["version", 8],
        ]

        parsed_prov_values = {}
        total_bit_width = sum([x[1] for x in prov_width_list])

        packet_in_reg = self.controller.read_register("MyIngress.prov_packet_in")
        # packet_in_reg = self.controller.read_register(
        #     "ingress.table0_control.prov_packet_in"
        # )
        # packet_in_reg = self.controller.read_register("prov_packet_in")
        prov_reg_indices = self.controller.read_register("MyIngress.prov_index_reg")
        # prov_reg_indices = self.controller.read_register(
        #     "ingress.table0_control.prov_index_reg"
        # )
        # prov_reg_indices = self.controller.read_register("prov_index_reg")
        end_idx = prov_reg_indices[index_map["packet_in"]]
        start_idx = self.last_count_index["packet_in"]
        idx = start_idx
        entries_read = 0
        # for prov_entry in packet_in_reg[self.last_count_index["packet_in"] :]:
        while idx != end_idx:
            # print("prov entry: ", prov_entry)
            prov_entry = packet_in_reg[idx]
            if prov_entry == 0:
                print(f"Entry 0 - PacketIn: idx: {idx} and end_idx:{end_idx}")
                break
            start_index = 0
            prov_string = format(prov_entry, "b").zfill(total_bit_width)
            # print(f"packetin bit string: ", prov_string)
            for entity_field in prov_width_list:
                entity_type = entity_field[0]
                entity_size = entity_field[1]
                end_index = start_index + entity_size
                entity_value = prov_string[start_index:end_index]
                if entity_type in PKT_IN_TYPE:
                    if PKT_IN_TYPE[entity_type] == "hex":
                        parsed_prov_values[entity_type] = hex(int(entity_value, 2))
                    elif PKT_IN_TYPE[entity_type] == "ip_addr":
                        parsed_prov_values[entity_type] = ".".join(
                            [
                                str((int(entity_value, 2) >> (8 * i)) & 255)
                                for i in range(3, -1, -1)
                            ]
                        )
                    elif PKT_IN_TYPE[entity_type] == "mac_addr":
                        parsed_prov_values[entity_type] = "{:04x}".format(
                            int(entity_value, 2)
                        )
                        parsed_prov_values[entity_type] = ":".join(
                            [
                                parsed_prov_values[entity_type][i : i + 2]
                                for i in range(0, 12, 2)
                            ]
                        )
                    else:
                        logger.warning("PKT_IN_TYPE not recognized")
                else:
                    # Default value type is INT
                    parsed_prov_values[entity_type] = int(entity_value, 2)
                start_index = end_index
                entries_read += 1
                idx = (idx + 1) % REGISTER_SIZE
            logger.info(f"PacketIn values: {parsed_prov_values}")
            self._prepare_prov_api(
                "packet_in", parsed_prov_values
            )  # Send the prov events to the queue
        self.last_count_index["packet_in"] = end_idx
        print(
            f"[INFO] Pulling PacketIN provenance.. {round(self.last_count_index['packet_in'] / REGISTER_SIZE, 3) * 100}"
        )
        logger.info(
            f"[INFO] Provenance register capacity: {self.last_count_index['packet_in'] / 65535}"
        )

    def _read_packet_out_prov(self):
        # TODO - Merge the packet_in and packet_out functions and use a flag

        prov_width_list = [
            ["src_mac", 16],
            ["packet_id", 16],
            ["egress_port", 8],
            ["ether_type", 16],
            ["version", 3],
        ]

        parsed_prov_values = {}
        total_bit_width = sum([x[1] for x in prov_width_list])

        packet_out_reg = self.controller.read_register(
            "ingress.firewall_control.prov_packet_out"
        )
        prov_reg_indices = self.controller.read_register(
            "ingress.firewall_control.prov_index_reg"
        )
        end_idx = prov_reg_indices[index_map["packet_out"]]
        start_idx = self.last_count_index["packet_out"]
        idx = start_idx
        entries_read = 0

        # for prov_entry in packet_in_reg[self.last_count_index["packet_out"] :]:
        while idx != end_idx:
            # print("prov entry: ", prov_entry)
            prov_entry = packet_out_reg[idx]
            if prov_entry == 0:
                print(f"Entry 0 - PacketIn: idx: {idx} and end_idx:{end_idx}")
                break
            # if prov_entry != 0:
            start_index = 0
            prov_string = format(prov_entry, "b").zfill(total_bit_width)
            # print(f"packetin bit string: ", prov_string)
            for entity_field in prov_width_list:
                entity_type = entity_field[0]
                entity_size = entity_field[1]
                end_index = start_index + entity_size
                entity_value = prov_string[start_index:end_index]
                if entity_type in PKT_IN_TYPE:
                    if PKT_IN_TYPE[entity_type] == "hex":
                        parsed_prov_values[entity_type] = hex(int(entity_value, 2))
                    elif PKT_IN_TYPE[entity_type] == "ip_addr":
                        parsed_prov_values[entity_type] = ".".join(
                            [
                                str((int(entity_value, 2) >> (8 * i)) & 255)
                                for i in range(3, -1, -1)
                            ]
                        )
                    elif PKT_IN_TYPE[entity_type] == "mac_addr":
                        parsed_prov_values[entity_type] = "{:04x}".format(
                            int(entity_value, 2)
                        )
                        parsed_prov_values[entity_type] = ":".join(
                            [
                                parsed_prov_values[entity_type][i : i + 2]
                                for i in range(0, 12, 2)
                            ]
                        )
                    else:
                        logger.warning("PktOut TYPE not recognized")
                else:
                    # Default value type is INT
                    parsed_prov_values[entity_type] = int(entity_value, 2)
                start_index = end_index
                entries_read += 1
                idx = (idx + 1) % REGISTER_SIZE
            logger.info(f"PacketOut values: {parsed_prov_values}")
            self._prepare_prov_api(
                "packet_out", parsed_prov_values
            )  # Send the prov events to the queue
            # else:
            self.last_count_index["packet_out"] = end_idx
            print(
                f"[INFO] Pulling PacketOut provenance.. {round(self.last_count_index['packet_out'] / 65535, 3) * 100}"
            )
            # logger.info(
            #     f"[INFO] Provenance register capacity: {self.last_count_index['packet_out'] / 65535}"
            # )

    def _prepare_prov_api(self, entity_name: str, entity_call_values: dict):
        activity_type = "PacketProcessor"
        entity_operation = ""
        version = entity_call_values.pop("version")

        if entity_name == "packet_in":
            # activity_id_str = str(entity_call_values["activity"])
            # activity_name = self.reg_prov_details["packet_in"][activity_id_str]
            activity_name = "send_to_cpu"
            # entity_call_values.pop("activity")
            # entity_call_values.update({"activity_name": activity_name})
            entity_operation = "WRITE"
        elif entity_name == "packet_out":
            activity_name = "set_egress_port"
            entity_call_values.update({"activity_name": activity_name})
            entity_operation = "READ"
        elif entity_name == "table_rule":
            logger.info("table rule API")
            activity_name = "set_egress_port"
        else:
            logger.error("Entity name not recognized...")
        # Use the ProvP4 API calls
        record_agent(self.switch_name, self.agent_id, "Switch", 0)
        record_activity(activity_name, self.agent_id, version, activity_type)
        record_API_call(
            entity_name,
            self.agent_id,
            version,
            entity_name,
            entity_operation,
            {"value": entity_call_values},
        )

    def start(self) -> None:
        """Start the collection of prov events. First prov registers then prov packetIn"""
        POLL_TIME = 2  # In seconds
        try:
            while True:
                self._prov_wraparound()
                # self._read_prov_regs()
                # self._read_packet_in_prov()
                # self._read_packet_out_prov()
                time.sleep(POLL_TIME)
        except KeyboardInterrupt:
            print("Closing DP collector")
            close_channel()


def main():
    """This function takes the provenance details, extracts the DP provenance, parses it
    and sends it to the RabbitMQ for processing using SPADE"""
    parser = argparse.ArgumentParser(
        description="This script requires the provenance detail file as input."
    )
    parser.add_argument("prov_details", type=str, help="Provenance details JSON")
    parser.add_argument("switch_name", type=str, help="Switch Name")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-f", "--thrift_port_file", type=str, help="Path to the thrift port file."
    )
    group.add_argument("-p", "--thrift_port", type=int, help="Thrift port directly.")
    args = parser.parse_args()

    if args.thrift_port_file:
        with open(args.thrift_port_file, "r") as file:
            line = file.readline().strip()
            thrift_port = int(line)
    elif args.thrift_port is not None:
        thrift_port = args.thrift_port

    args = parser.parse_args()
    with open(args.prov_details, "r") as file:
        reg_prov_details = json.load(file)
    collector = ProvCollector(reg_prov_details, thrift_port, args.switch_name)
    collector.start()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import logging
import os
import re
import time
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Union

import requests
import yaml
from atomicwrites import atomic_write

from fabrictestbed_extensions import __version__ as fablib_version
from fabrictestbed_extensions.fablib.constants import Constants
from fabrictestbed_extensions.utils.utils import Utils


class ConfigException(Exception):
    """
    An exception class to represent configuration errors.
    """

    pass


class Config:
    """
    A class that represents fablib configuration.
    """

    LOG_LEVELS = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    # Mapping of the required parameters and the corresponding Environment Variable name
    REQUIRED_ATTRS = {
        Constants.ORCHESTRATOR_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_ORCHESTRATOR_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_ORCHESTRATOR_HOST,
        },
        Constants.CREDMGR_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_CREDMGR_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_CREDMGR_HOST,
        },
        Constants.CORE_API_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_CORE_API_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_CORE_API_HOST,
        },
        Constants.AM_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_AM_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_AM_HOST,
        },
        Constants.TOKEN_LOCATION: {
            Constants.ENV_VAR: Constants.FABRIC_TOKEN_LOCATION,
            Constants.DEFAULT: Constants.DEFAULT_TOKEN_LOCATION,
        },
        Constants.PROJECT_ID: {
            Constants.ENV_VAR: Constants.FABRIC_PROJECT_ID,
        },
        Constants.BASTION_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_BASTION_HOST,
        },
        Constants.BASTION_USERNAME: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_USERNAME,
        },
        Constants.BASTION_KEY_LOCATION: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_KEY_LOCATION,
            Constants.DEFAULT: Constants.DEFAULT_BASTION_KEY_LOCATION,
        },
        Constants.FABLIB_VERSION: {Constants.DEFAULT: fablib_version},
        Constants.SLICE_PUBLIC_KEY_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_SLICE_PUBLIC_KEY_FILE,
            Constants.DEFAULT: Constants.DEFAULT_SLICE_PUBLIC_KEY_FILE,
        },
        Constants.SLICE_PRIVATE_KEY_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_SLICE_PRIVATE_KEY_FILE,
            Constants.DEFAULT: Constants.DEFAULT_SLICE_PRIVATE_KEY_FILE,
        },
        Constants.AVOID: {
            Constants.ENV_VAR: Constants.FABRIC_AVOID,
            Constants.DEFAULT: "",
        },
        Constants.SSH_COMMAND_LINE: {
            Constants.ENV_VAR: Constants.FABRIC_SSH_COMMAND_LINE,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_SSH_COMMAND_LINE,
        },
        Constants.LOG_LEVEL: {
            Constants.ENV_VAR: Constants.FABRIC_LOG_LEVEL,
            Constants.DEFAULT: Constants.DEFAULT_LOG_LEVEL,
        },
        Constants.LOG_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_LOG_FILE,
            Constants.DEFAULT: Constants.DEFAULT_LOG_FILE,
        },
        Constants.BASTION_SSH_CONFIG_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_SSH_CONFIG_FILE,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_BASTION_SSH_CONFIG_FILE,
        },
        Constants.METADATA_TAG: {
            Constants.ENV_VAR: Constants.FABRIC_METADATA_TAG,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_METADATA_TAG,
        },
    }

    REQUIRED_ATTRS_PRETTY_NAMES = {
        Constants.CREDMGR_HOST: "Credential Manager",
        Constants.ORCHESTRATOR_HOST: "Orchestrator",
        Constants.CORE_API_HOST: "Core API",
        Constants.AM_HOST: "Artifact Manager",
        Constants.TOKEN_LOCATION: "Token File",
        Constants.PROJECT_ID: "Project ID",
        Constants.PROJECT_NAME: "Project Name",
        Constants.BASTION_USERNAME: "Bastion Username",
        Constants.BASTION_KEY_LOCATION: "Bastion Private Key File",
        Constants.BASTION_HOST: "Bastion Host",
        Constants.BASTION_KEY_PASSPHRASE: "Bastion Private Key Passphrase",
        Constants.SLICE_PUBLIC_KEY_FILE: "Slice Public Key File",
        Constants.SLICE_PRIVATE_KEY_FILE: "Slice Private Key File",
        Constants.SLICE_PRIVATE_KEY_PASSPHRASE: "Slice Private Key Passphrase",
        Constants.LOG_FILE: "Log File",
        Constants.LOG_LEVEL: "Log Level",
        Constants.FABLIB_VERSION: "Version",
        Constants.AVOID: "Sites to avoid",
        Constants.DATA_DIR: "Data directory",
        Constants.SSH_COMMAND_LINE: "SSH Command Line",
        Constants.BASTION_SSH_CONFIG_FILE: "Bastion SSH Config File",
        Constants.METADATA_TAG: "Fabric Meta Data Release Tag",
    }

    os.makedirs(Constants.LOCAL_CACHE_DIR, exist_ok=True)

    def __init__(
        self,
        fabric_rc: str = None,
        credmgr_host: str = None,
        orchestrator_host: str = None,
        core_api_host: str = None,
        am_host: str = None,
        token_location: str = None,
        project_id: str = None,
        bastion_username: str = None,
        bastion_key_location: str = None,
        log_level: str = Constants.DEFAULT_LOG_LEVEL,
        log_file: str = Constants.DEFAULT_LOG_FILE,
        data_dir: str = Constants.DEFAULT_DATA_DIR,
        offline: bool = True,
        **kwargs,
    ):
        """
        Constructor. Tries to get configuration from:

         - constructor parameters (high priority)
         - fabric_rc file (middle priority)
         - environment variables (low priority)
         - defaults (if needed and possible)

        """
        if fabric_rc is None:
            fabric_rc = Constants.DEFAULT_FABRIC_RC
            if os.path.exists(Constants.DEFAULT_FABRIC_CONFIG_DIR):
                Path(fabric_rc).touch()

        if fabric_rc and os.path.exists(fabric_rc):
            self.config_file_path = fabric_rc
        self.is_yaml = False
        self.runtime_config = {}
        self.offline = offline

        # Load from config file
        self.__load_configuration(file_path=fabric_rc, **kwargs)

        # Apply any parameters explicitly passed
        if credmgr_host is not None:
            self.set_credmgr_host(credmgr_host=credmgr_host)

        if orchestrator_host is not None:
            self.set_orchestrator_host(orchestrator_host=orchestrator_host)

        if core_api_host is not None:
            self.set_core_api_host(core_api_host=core_api_host)

        if am_host is not None:
            self.set_am_host(am_host=am_host)

        if token_location is not None:
            self.set_token_location(token_location=token_location)

        if project_id is not None:
            self.set_project_id(project_id=project_id)

        if bastion_username is not None:
            self.set_bastion_username(bastion_username=bastion_username)

        if bastion_key_location is not None:
            self.set_bastion_key_location(bastion_key_location=bastion_key_location)

        if log_level is not None:
            self.set_log_level(log_level=log_level)

        if log_file is not None:
            self.set_log_file(log_file=log_file)

        if data_dir is not None:
            self.set_data_dir(data_dir=data_dir)

        #        if self.get_ssh_command_line() is None:
        #            self.set_ssh_command_line(
        #                ssh_command_line=Constants.DEFAULT_FABRIC_SSH_COMMAND_LINE
        #            )

        self.required_check(partial=True)

        # Verify that Token file exists; any other checks cannot be done without this.
        token_location = self.get_token_location()
        if not os.path.exists(token_location):
            raise ConfigException(
                f"Token file does not exist, please provide the token at location: {token_location}!"
            )

    def __load_configuration(self, file_path, **kwargs):
        """
        Load the config parameters from config file;
        - Parses bashrc/yaml file and loads the parameters in dictionary

        :param file_path: path to the config file; can be bashrc or yaml
        :type file_path: str
        :param kwargs:
        :type kwargs:

        :raises ConfigException: if config file does not exist
        """
        if file_path and os.path.exists(file_path):
            if Utils.is_yaml_file(file_path=file_path):
                self.__load_yaml_file(file_path=file_path)
                self.is_yaml = True
            else:
                self.__load_fabric_rc_file(file_path=file_path)

        for k, v in kwargs.items():
            self.runtime_config[k] = v

    def __load_yaml_file(self, file_path):
        """
        Load the config from yml file
        :param file_path: path to the config file
        :type file_path: str

        :return True or False indicating success/failure for loading the config file
        :rtype bool
        """
        try:
            with open(file_path, "r") as file:
                config = yaml.safe_load(file)
                if (
                    isinstance(config, dict)
                    and config.get(Constants.RUNTIME_SECTION) is not None
                ):
                    for key, value in config.get(Constants.RUNTIME_SECTION).items():
                        if value is not None:
                            value = value.strip().strip("'").strip('"')
                            if key == Constants.AVOID:
                                value = value.split(",")
                            self.runtime_config[key] = value
                    self.runtime_config[Constants.FABLIB_VERSION] = fablib_version
                    return True
                return False
        except yaml.YAMLError:
            return False

    def __load_fabric_rc_file(self, file_path):
        """
        Load the config from bashrc file
        :param file_path: path to the config file
        :type file_path: str

        :return True or False indicating success/failure for loading the config file
        :rtype bool
        """
        ret_val = False
        with open(file_path, "r") as file:
            lines = file.readlines()

        export_pattern = re.compile(r"^export\s+([^=]+)=(.*)$", re.IGNORECASE)

        for line in lines:
            match = export_pattern.match(line.strip())
            if match:
                key, value = match.groups()
                value = value.strip()
                value = value.strip('"')
                value = value.strip("'")
                key = key.replace("FABRIC_", "").lower()
                if key == Constants.AVOID and value is not None and len(value):
                    value = value.split(",")
                self.runtime_config[key] = value
                ret_val = True
        self.runtime_config[Constants.FABLIB_VERSION] = fablib_version
        return ret_val

    def required_check(self, partial: bool = False):
        """
        Validates that all the required parameters are present in the loaded config
        Tries to load the parameters from environment variable or defauls if not available in config in that order

        :raises AttributeError if missing required parameters are found
        """
        errors = []

        for attr, attr_props in self.REQUIRED_ATTRS.items():
            if attr not in self.runtime_config or self.runtime_config.get(attr) is None:
                # Load from environment variables
                if (
                    attr_props.get(Constants.ENV_VAR)
                    and os.environ.get(attr_props.get(Constants.ENV_VAR)) is not None
                ):
                    self.runtime_config[attr] = os.environ.get(
                        attr_props.get(Constants.ENV_VAR)
                    )
                # Load defaults if available
                elif attr_props.get(Constants.DEFAULT) is not None:
                    self.runtime_config[attr] = attr_props.get(Constants.DEFAULT)
                # Ignore validation for this parameter if no default value exists and this partial check
                elif attr_props.get(Constants.DEFAULT) is None and partial:
                    continue
                else:
                    errors.append(f"{attr} is not set")

        if errors:
            logging.error(f"Failing Config: {self.runtime_config}")
            # TODO: define custom exception class to report errors,
            # and emit a more helpful error message with hints about
            # setting up environment variables or configuration file.
            raise AttributeError(
                f"Error initializing {self.__class__.__name__}: {errors}"
            )

    def get_config(self) -> Dict[str, str]:
        """
        Gets a dictionary mapping keywords to configured FABRIC environment
        variable values.

        :return: dictionary mapping keywords to FABRIC values
        :rtype: Dict[String, String]
        """
        return self.runtime_config

    def get_default_slice_public_key_file(self) -> str:
        """
        Gets the path to the default slice public key file.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice public key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PUBLIC_KEY_FILE)

    def get_default_slice_private_key_file(self) -> str:
        """
        Gets the path to the default slice private key file.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice private key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PRIVATE_KEY_FILE)

    def get_default_slice_private_key_passphrase(self) -> str:
        """
        Gets the passphrase to the default slice private key.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the passphrase to the slice private key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PRIVATE_KEY_PASSPHRASE)

    def get_credmgr_host(self) -> str:
        """
        Gets the credential manager host value.

        :return: the credential manager host
        :rtype: String
        """
        return self.get_config().get(Constants.CREDMGR_HOST)

    def set_credmgr_host(self, credmgr_host: str):
        """
        Sets the credential manager host value.
        :param credmgr_host: credential manager host site
        :type credmgr_host: String
        """
        self.runtime_config[Constants.CREDMGR_HOST] = credmgr_host

    def get_orchestrator_host(self) -> str:
        """
        Gets the orchestrator host value.

        :return: the orchestrator host site
        :rtype: String
        """
        return self.get_config().get(Constants.ORCHESTRATOR_HOST)

    def set_orchestrator_host(self, orchestrator_host: str):
        """
        Sets the Orchestrator host value.
        :param orchestrator_host: Orchestrator host
        :type orchestrator_host: String
        """
        self.runtime_config[Constants.ORCHESTRATOR_HOST] = orchestrator_host

    def get_core_api_host(self) -> str:
        """
        Gets the core_api host value.

        :return: the core_api host site
        :rtype: String
        """
        return self.get_config().get(Constants.CORE_API_HOST)

    def set_core_api_host(self, core_api_host: str):
        """
        Sets the core_api host value.
        :param core_api_host: core_api host
        :type core_api_host: String
        """
        self.runtime_config[Constants.CORE_API_HOST] = core_api_host

    def get_am_host(self) -> str:
        """
        Gets the am host value.

        :return: the am host site
        :rtype: String
        """
        return self.get_config().get(Constants.AM_HOST)

    def set_am_host(self, am_host: str):
        """
        Sets the am host value.
        :param am_host: core_api host
        :type am_host: String
        """
        self.runtime_config[Constants.AM_HOST] = am_host

    def get_token_location(self) -> str:
        """
        Gets the FABRIC token location.

        :return: FABRIC token location
        :rtype: String
        """
        return self.get_config().get(Constants.TOKEN_LOCATION)

    def set_token_location(self, token_location: str):
        """
        Sets the FABRIC token location.

        :param token_location: Token Location
        :type token_location: String
        """
        self.runtime_config[Constants.TOKEN_LOCATION] = token_location

    def get_bastion_username(self) -> str:
        """
        Gets the FABRIC Bastion username.

        :return: FABRIC Bastion username
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_USERNAME)

    def set_bastion_username(self, bastion_username: str):
        """
        Sets the FABRIC Bastion username.

        :param bastion_username: Bastion username
        :type bastion_username: String
        """
        self.runtime_config[Constants.BASTION_USERNAME] = bastion_username

    def get_bastion_key_location(self) -> str:
        """
        Gets the FABRIC Bastion key filename.

        :return: FABRIC Bastion key filename
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_KEY_LOCATION)

    def set_bastion_key_location(self, bastion_key_location: str):
        """
        Sets the FABRIC Bastion Key Location.

        :param bastion_key_location: Bastion Key Location
        :type bastion_key_location: String
        """
        self.runtime_config[Constants.BASTION_KEY_LOCATION] = bastion_key_location

    def get_bastion_key_passphrase(self) -> str:
        """
        Reads the FABRIC Bastion private key passphrase.

        :return: FABRIC Bastion key passphrase
        :rtype: String
        """
        return self.runtime_config.get(Constants.BASTION_KEY_PASSPHRASE)

    def get_bastion_key(self) -> Union[str, None]:
        """
        Reads the FABRIC Bastion private key file and returns the key.

        :return: FABRIC Bastion key string
        :rtype: String
        """
        if self.get_bastion_key_location() is None or not os.path.exists(
            self.get_bastion_key_location()
        ):
            return None
        return Utils.read_file_contents(file_path=self.get_bastion_key_location())

    def get_bastion_public_key(self) -> Union[str, None]:
        """
        Reads the FABRIC Bastion public key file and returns the key.

        :return: FABRIC Bastion key string
        :rtype: String
        """
        if self.get_bastion_key_location() is None or not os.path.exists(
            self.get_bastion_key_location()
        ):
            return None
        return Utils.read_file_contents(
            file_path=f"{self.get_bastion_key_location()}.pub"
        )

    def get_bastion_host(self) -> str:
        """
        Gets the FABRIC Bastion host address.

        :return: Bastion host public address
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_HOST)

    def get_project_id(self) -> str:
        """
        Get the Project Id:
        :return: Project Id
        :rtype: String
        """
        return self.get_config().get(Constants.PROJECT_ID)

    def set_project_id(self, project_id: str):
        """
        Set the Project Id:
        :param: project_id: Project Id
        :type: String
        """
        self.runtime_config[Constants.PROJECT_ID] = project_id

    def get_project_name(self):
        """
        Get the Project Name:
        :return: Project Name
        :rtype: String
        """
        return self.runtime_config.get(Constants.PROJECT_NAME)

    def set_log_level(self, log_level: str = "INFO"):
        """
        Sets the current log level for logging

        Options:  'DEBUG'
                  'INFO'
                  'WARNING'
                  'ERROR'
                  'CRITICAL'

        :param log_level: new log level
        :type str: Level
        """

        self.runtime_config[Constants.LOG_LEVEL] = log_level

    def get_log_level(self):
        """
        Get the current log level for logging

        :return log_level: new log level
        :rtype log_level: string
        """

        return self.runtime_config.get(Constants.LOG_LEVEL)

    def get_log_file(self) -> str:
        """
        Gets the current log file for logging

        :return log_file: new log level
        :rtype log_file: string
        """

        return self.runtime_config.get(Constants.LOG_FILE)

    def set_log_file(self, log_file: str):
        """
        Sets the current log file for logging

        :param log_file: new log level
        :type log_file: string
        """
        self.runtime_config[Constants.LOG_FILE] = log_file

    def get_data_dir(self) -> str:
        """
        Gets the data directory

        :return data_dir: data directory
        :rtype data_dir: string
        """

        return self.runtime_config.get(Constants.DATA_DIR)

    def set_data_dir(self, data_dir: str):
        """
        Sets the data directory

        :param data_dir: data directory
        :type data_dir: string
        """
        self.runtime_config[Constants.DATA_DIR] = data_dir

    def get_ssh_command_line(self):
        """
        Gets the data directory

        :return data_dir: data directory
        :rtype data_dir: string
        """
        return self.runtime_config.get(Constants.SSH_COMMAND_LINE)

    def set_ssh_command_line(self, ssh_command_line: str):
        """
        Sets the ssh command line

        :param ssh_command_line: ssh command line
        :type ssh_command_line: string
        """
        self.runtime_config[Constants.SSH_COMMAND_LINE] = ssh_command_line

    def get_bastion_ssh_config_file(self):
        """
        Gets the bastion ssh config file

        :return bastion_ssh_config_file: bastion_ssh_config_file
        :rtype bastion_ssh_config_file: string
        """
        return self.runtime_config.get(Constants.BASTION_SSH_CONFIG_FILE)

    def get_avoid(self):
        """
        Gets the avoid list

        :return avoid: avoid list
        :rtype avoid: string
        """
        return self.runtime_config.get(Constants.AVOID)

    def set_avoid(self, avoid: List[str]):
        """
        Sets the avoid list

        :param avoid: avoid list
        :type avoid: string
        """
        self.runtime_config[Constants.AVOID] = avoid

    def set_avoid_csv(self, avoid_csv: str = ""):
        """
        Sets the avoid site list from a CSV

        :param avoid_csv: avoid list
        :type avoid_csv: List of string
        """
        avoid_csv = avoid_csv.strip().strip('"').strip("'")

        avoid = []
        for site in avoid_csv.split(","):
            avoid.append(site.strip())

        self.set_avoid(avoid)

    def get_default_slice_private_key(self) -> Union[str, None]:
        """
        Gets the current default_slice_keys as a dictionary containg the
        public and private slice keys.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: default_slice_key dictionary from superclass
        :rtype: Dict[String, String]
        """
        if self.get_default_slice_private_key_file() is not None and os.path.exists(
            self.get_default_slice_private_key_file()
        ):
            return Utils.read_file_contents(
                file_path=self.get_default_slice_private_key_file()
            )
        return None

    def get_default_slice_public_key(self) -> Union[str, None]:
        """
        Gets the current default_slice_keys as a dictionary containg the
        public and private slice keys.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: default_slice_key dictionary from superclass
        :rtype: Dict[String, String]
        """
        if self.get_default_slice_public_key_file() is not None and os.path.exists(
            self.get_default_slice_public_key_file()
        ):
            return Utils.read_file_contents(
                file_path=self.get_default_slice_public_key_file()
            )
        return None

    @staticmethod
    def get_image_names() -> dict[str, dict]:
        """
        Gets a list of available image names.

        This is statically defined for now. Eventually, images will be managed dynamically.

        :return: Dictionary of images with default user and description
        :rtype: dict[str, dict]
        """
        return Constants.IMAGE_NAMES

    def get_config_pretty_names_dict(self) -> Dict[str, str]:
        """
        Return PRETTY Names for the config

        :return: Dict of Pretty Names
        :rtype: Dict[str, str]
        """
        return self.REQUIRED_ATTRS_PRETTY_NAMES

    def save_config(self):
        """
        Write the configuration file.
        """
        if self.config_file_path is None:
            print("Config file path not set!")
            return

        if self.is_yaml:
            # Write the dictionary to the YAML file
            with atomic_write(self.config_file_path, overwrite=True) as f:
                yaml.dump(self.runtime_config, f, default_flow_style=False)
        else:
            with atomic_write(self.config_file_path, overwrite=True) as f:
                for attr, attr_props in self.REQUIRED_ATTRS.items():
                    env_var = attr_props.get(Constants.ENV_VAR)
                    value = self.runtime_config.get(attr)
                    if value is None:
                        value = ""
                    if env_var:
                        f.write(f"export {env_var}={value}\n")

    def setup_logging(self):
        """
        Create log file if it doesn't exist; setup logger
        """
        try:
            for handler in logging.root.handlers[:]:
                logging.root.removeHandler(handler)
        except Exception as e:
            print(f"Exception from removeHandler: {e}")
            pass

        try:
            if self.get_log_file() and not os.path.isdir(
                os.path.dirname(self.get_log_file())
            ):
                os.makedirs(os.path.dirname(self.get_log_file()))
        except Exception:
            logging.warning(
                f"Failed to create log_file directory: {os.path.dirname(self.get_log_file())}"
            )

        if self.get_log_file() and self.get_log_level():
            logging.basicConfig(
                filename=self.get_log_file(),
                level=self.LOG_LEVELS[self.get_log_level()],
                format="[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s",
                datefmt="%H:%M:%S",
            )

    @staticmethod
    def get_metadata_tag() -> str:
        """
        Get the metadata tag used to fetch remote configuration files.

        This tag determines which version or branch (e.g., "main", "v1.0.0")
        of the metadata directory is used.

        :return: Metadata tag string.
        :rtype: str
        """
        return os.environ.get(
            Constants.FABRIC_METADATA_TAG, Constants.DEFAULT_FABRIC_METADATA_TAG
        )

    @staticmethod
    def _fetch_or_load_json(
        name: str, ttl: int = 86400, fallback_default: dict = {}
    ) -> dict:
        """
        Fetch a JSON file from a remote URL, using local caching with time-based invalidation,
        and fall back to a default if the fetch or cache read fails.

        If a valid cached file exists within the TTL, it is used.
        If not, the method tries to fetch from the remote GitHub metadata URL.
        If the fetch fails, it returns either the cached file (if available) or the fallback default.

        :param name: Name of the metadata file (without .json extension).
        :type name: str
        :param ttl: Time-to-live for cache in seconds (default is 86400 = 1 day).
        :type ttl: int
        :param fallback_default: Default dictionary to use if remote and cached fetches fail.
        :type fallback_default: dict or None
        :return: The loaded JSON data.
        :rtype: dict
        :raises RuntimeError: If no valid data can be fetched or loaded and no fallback is provided.
        """
        remote_url = Constants.FABRIC_METADATA_URL.format(Config.get_metadata_tag())
        remote_url += f"/{name}.json"
        local_file = os.path.join(Constants.LOCAL_CACHE_DIR, f"{name}.json")

        # Check if cache file is fresh
        if os.path.exists(local_file):
            age = time.time() - os.path.getmtime(local_file)
            if age < ttl:
                with open(local_file, "r") as f:
                    return json.load(f)

        try:
            response = requests.get(remote_url, timeout=5)
            response.raise_for_status()
            data = response.json()
            with atomic_write(local_file, overwrite=True) as f:
                json.dump(data, f, indent=2)
            return data
        except Exception:
            if os.path.exists(local_file):
                with open(local_file, "r") as f:
                    return json.load(f)
            elif fallback_default is not None:
                return fallback_default
            else:
                raise RuntimeError(f"Could not retrieve or load {name}.json")

    @staticmethod
    @lru_cache(maxsize=1)
    def get_os_images() -> dict:
        """
        Get a dictionary of supported OS images from cached metadata.

        The data is fetched from a remote GitHub metadata source or returned from local cache.
        Falls back to a static default (Constants.IMAGE_NAMES) if both fail.

        :return: Dictionary of image metadata.
        :rtype: dict
        """
        return Config._fetch_or_load_json(
            "os_images", fallback_default=Constants.IMAGE_NAMES
        )

    @staticmethod
    @lru_cache(maxsize=1)
    def get_facility_port_details() -> dict:
        """
        Get the facility port metadata including descriptions and mappings.

        Uses remote metadata cache if available, falling back to an empty dictionary.

        :return: Dictionary of facility port details.
        :rtype: dict
        """
        return Config._fetch_or_load_json("facility_ports", fallback_default={})


if __name__ == "__main__":
    config1 = Config(fabric_rc="./fabric_rc", credmgr_host="abc.def.com")
    assert config1.get_credmgr_host() == "abc.def.com"
    print(json.dumps(config1.runtime_config, indent=4))

    config2 = Config(fabric_rc="./fabric_rc.yml", credmgr_host="abc.def.com")
    assert config2.get_credmgr_host() == "abc.def.com"
    print(json.dumps(config2.runtime_config, indent=4))
